import collections
import queue
import multiprocessing
import os
import typing
import threading

import psutil

from ..easel import Alphabet, DigitalSequence, DigitalMSA, DigitalSequenceBlock, SequenceFile
from ..plan7 import TopHits, Builder, Pipeline, Background, HMM, Profile, OptimizedProfile, HMMFile, HMMPressedFile, OptimizedProfileBlock
from ..utils import singledispatchmethod
from ._base import _BaseDispatcher, _BaseWorker, _BaseChore, _AnyProfile

if typing.TYPE_CHECKING:
    from ._base import Unpack, PipelineOptions, BACKEND

# --- Worker -------------------------------------------------------------------

class _SCANWorker(
    _BaseWorker[
        DigitalSequence,
        typing.Union[OptimizedProfileBlock, HMMPressedFile],
        "TopHits[DigitalSequence]",
    ],
    threading.Thread
):
    @singledispatchmethod
    def query(self, query) -> "TopHits[Any]":  # type: ignore
        raise TypeError(
            "Unsupported query type for `hmmscan`: {}".format(type(query).__name__)
        )

    @query.register(DigitalSequence)
    def _(self, query: DigitalSequence) -> "TopHits[DigitalSequence]":  # type: ignore
        assert self.pipeline is not None
        return self.pipeline.scan_seq(query, self.targets)


class _SCANThread(_SCANWorker, threading.Thread):
    pass


class _SCANProcess(_SCANWorker, multiprocessing.Process):
    pass

# --- Dispatcher ---------------------------------------------------------------

class _SCANDispatcher(
    _BaseDispatcher[
        DigitalSequence,
        typing.Union[OptimizedProfileBlock, HMMPressedFile],
        "TopHits[DigitalSequence]",
    ]
):
    def _new_worker(
        self,
        query_queue: "queue.Queue[typing.Optional[_BaseChore[DigitalSequence, TopHits[DigitalSequence]]]]",
        query_count: "multiprocessing.Value[int]",  # type: ignore
        kill_switch: threading.Event,
    ) -> _SCANWorker:
        if isinstance(self.targets, HMMPressedFile):
            assert self.targets.name is not None
            targets = HMMPressedFile(self.targets.name)
        else:
            targets = self.targets  # type: ignore
        params = [
            targets,
            query_queue,
            query_count,
            kill_switch,
            self.callback,
            self.options,
        ]
        if self.backend == "threading":
            return _SCANThread(*params)
        elif self.backend == "multiprocessing":
            return _SCANProcess(*params)
        else:
            raise ValueError(f"Invalid backend for `hmmsearch`: {self.backend!r}")


# --- hmmscan ----------------------------------------------------------------

def hmmscan(
    queries: typing.Union[DigitalSequence, typing.Iterable[DigitalSequence]],
    profiles: typing.Iterable[typing.Union[HMM, Profile, OptimizedProfile]],
    *,
    cpus: int = 0,
    callback: typing.Optional[typing.Callable[[DigitalSequence, int], None]] = None,
    backend: "BACKEND" = "threading",
    **options,  # type: Unpack[PipelineOptions]
) -> typing.Iterator["TopHits[DigitalSequence]"]:
    """Scan query sequences against a profile database.

    In HMMER many-to-many comparisons, a *scan* is the operation of querying
    with sequences a database of profile HMMs. It is necessary slower than
    a *search* because reconfiguring profiles between each queries has
    additional overhead, so it's recommended to use a *search* if the order
    of the comparisons is not important.

    The `hmmscan` function offers two ways of managing the database that will
    be selected based on the type of the ``profiles`` argument. If
    ``profiles`` is an `HMMPressedFile` object, `hmmscan` will reopen the
    file in each thread, and load profiles *iteratively* to scan with the
    query. Otherwise, it will *pre-fetch* the optimized profiles into an
    `OptimizedProfileBlock` collection, and share them across queries. The
    *pre-fetching* gives much higher performance at the cost of extra
    startup time and much higher memory consumption. You may want to check
    how much memory is available (for instance with `psutil.virtual_memory`)
    before trying to load a whole pHMM database.

    Arguments:
        queries (iterable of `DigitalSequence`): The query sequences to scan
            with the database. Passing a single query is supported.
        profiles (iterable of `HMM`, `Profile` or `OptimizedProfile`): A
            database of profiles to query. If you plan on using the
            same targets several times, consider converting them into
            `OptimizedProfile` and storing them into an `OptimizedProfileBlock`
            ahead of time. If a `HMMPressedFile` is given, profiles will be
            loaded iteratively from disk rather than prefetched.
        cpus (`int`): The number of threads to run in parallel. Pass ``1``
            to run everything in the main thread, ``0`` to automatically
            select a suitable number (using `psutil.cpu_count`), or any
            positive number otherwise.
        callback (callable): A callback that is called everytime a query is
            processed with two arguments: the query, and the total number
            of queries. This can be used to display progress in UI.
        background (`pyhmmer.plan7.Background`, *optional*): A background
            object to use for configuring the profiles. If `None` given,
            create a default one.
        backend (`str`): The parallel backend to use for workers to be
            executed. Supports ``threading`` to use thread-based parallelism,
            or ``multiprocessing`` to use process-based parallelism.

    Yields:
        `~pyhmmer.plan7.TopHits`: An object reporting *top hits* for each
        query, in the same order the queries were passed in the input.

    Raises:
        `~pyhmmer.errors.AlphabetMismatch`: When any of the query sequence
            and the profile do not share the same alphabet.

    Note:
        Any additional keyword arguments passed to the `phmmer` function
        will be passed transparently to the `~pyhmmer.plan7.Pipeline` to
        be created in each worker thread.

    Hint:
        If reading the profiles from a pressed HMM database, make sure to
        use the `HMMFile.optimized_profiles` method so that profiles are
        read iteratively from the file during the scan loop::

            >>> with HMMFile("tests/data/hmms/db/PF02826.hmm") as hmm_file:
            ...     targets = hmm_file.optimized_profiles()
            ...     all_hits = list(hmmscan(proteins, targets, E=1e-10, cpus=1))
            >>> sum(len(hits) for hits in all_hits)
            6

        Otherwise, passing ``hmm_file`` as the ``profiles`` argument of
        `hmmscan` would cause the entire HMM file to be loaded in memory
        into an `OptimizedProfileBlock` otherwise.

    .. versionadded:: 0.7.0

    """
    cpus = cpus if cpus > 0 else psutil.cpu_count(logical=False) or os.cpu_count() or 1
    alphabet = options.get("alphabet")
    background = options.get("background")

    if not isinstance(queries, collections.abc.Iterable):
        queries = (queries,)
    if isinstance(profiles, HMMPressedFile):
        opt = profiles.read()
        profiles.rewind()
        if opt is not None:
            alphabet = alphabet or opt.alphabet
        else:
            alphabet = Alphabet.amino()
        targets = profiles
    elif isinstance(profiles, OptimizedProfileBlock):
        alphabet = alphabet or profiles.alphabet
        targets = profiles  # type: ignore
    else:
        block = None
        for item in profiles:
            alphabet = alphabet or item.alphabet
            if block is None:
                block = OptimizedProfileBlock(item.alphabet)
            if isinstance(item, HMM):
                if background is None:
                    background = Background(item.alphabet)
                profile = Profile(item.M, item.alphabet)
                profile.configure(item, background)
                item = profile
            if isinstance(item, Profile):
                item = item.to_optimized()
            if isinstance(item, OptimizedProfile):
                block.append(item)
            else:
                ty = type(item).__name__
                raise TypeError(
                    "Expected HMM, Profile or OptimizedProfile, found {}".format(ty)
                )
        if alphabet is None:
            alphabet = Alphabet.amino()
        if block is None:
            block = OptimizedProfileBlock(alphabet)
        targets = block  # type: ignore

    if "alphabet" not in options:
        options["alphabet"] = alphabet
    if "background" not in options and background is not None:
        options["background"] = background
    dispatcher = _SCANDispatcher(
        queries=queries,
        targets=targets,
        cpus=cpus,
        backend=backend,
        callback=callback,
        builder=None,
        **options,
    )
    return dispatcher.run()




from libc.stdint cimport uint64_t
from libc.stdio cimport FILE


cdef extern from "hmmer.h" nogil:

    const size_t p7G_NSCELLS
    cdef enum p7g_scells_e:
        p7G_M = 0
        p7G_I = 1
        p7G_D = 2

    const size_t p7G_NXCELLS
    cdef enum p7g_xcells_e:
        p7G_E  = 0
        p7G_N  = 1
        p7G_J  = 2
        p7G_B  = 3
        p7G_C  = 4

    ctypedef p7_gmx_s P7_GMX
    cdef struct p7_gmx_s:
        int  M
        int  L
        int      allocR
        int      validR
        int      allocW
        uint64_t ncells
        float **dp
        float  *xmx
        float  *dp_mem

    P7_GMX *p7_gmx_Create (int allocM, int allocL)
    int     p7_gmx_GrowTo (P7_GMX *gx, int allocM, int allocL)
    size_t  p7_gmx_Sizeof (P7_GMX *gx)
    int     p7_gmx_Reuse  (P7_GMX *gx)
    void    p7_gmx_Destroy(P7_GMX *gx)
    int     p7_gmx_Compare(P7_GMX *gx1, P7_GMX *gx2, float tolerance)
    int     p7_gmx_Dump(FILE *fp, P7_GMX *gx, int flags)
    int     p7_gmx_DumpWindow(FILE *fp, P7_GMX *gx, int istart, int iend, int kstart, int kend, int show_specials)



import copy
import io
import itertools
import os
import pickle
import platform
import shutil
import sys
import tempfile
import unittest

import pyhmmer
from pyhmmer.errors import EaselError, AlphabetMismatch
from pyhmmer.easel import Alphabet, SequenceFile, VectorF, TextSequence, Randomness
from pyhmmer.plan7 import HMM, HMMFile, Pipeline

from .. import __name__ as __package__
from .utils import HMMER_FOLDER, resource_files


class TestHMM(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        abc = Alphabet.dna()
        rng = Randomness()
        cls.hmm = HMM.sample(abc, 100, rng)

    @unittest.skipUnless(resource_files, "importlib.resources.files not available")
    def test_checksum(self):
        abc = Alphabet.dna()
        hmm = HMM(abc, 100, b"test")
        self.assertIs(hmm.checksum, None)

        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "LuxC.hmm")
        with HMMFile(hmm_path) as hmm_file:
            hmm = next(hmm_file)
        self.assertEqual(hmm.checksum, 395769323)

    def test_composition(self):
        abc = Alphabet.dna()
        hmm = HMM(abc, 100, b"test")
        self.assertIs(hmm.composition, None)
        hmm.set_composition()
        self.assertEqual(len(hmm.composition), hmm.alphabet.K)
        self.assertEqual(hmm.composition.shape, (hmm.alphabet.K,))

    def test_copy(self):
        hmm = copy.copy(self.hmm)
        self.assertEqual(self.hmm.name, hmm.name)
        self.assertEqual(self.hmm.accession, hmm.accession)
        self.assertEqual(self.hmm.description, hmm.description)
        self.assertEqual(self.hmm.consensus, hmm.consensus)

    def test_eq(self):
        abc = Alphabet.amino()
        self.assertEqual(self.hmm, self.hmm)
        self.assertEqual(self.hmm, self.hmm.copy())
        self.assertNotEqual(self.hmm, HMM(abc, 100, b"test"))
        self.assertNotEqual(self.hmm, 1)

    def test_insert_emissions(self):
        emissions = self.hmm.insert_emissions
        self.assertEqual(emissions.shape, (self.hmm.M+1, self.hmm.alphabet.K))

        row = emissions[0]
        self.assertAlmostEqual(row.sum(), 1.0, places=5)

    def test_command_line(self):
        abc = Alphabet.amino()
        hmm = HMM(abc, 100, b"test")
        self.assertIs(hmm.command_line, None)
        hmm.command_line = "hmmbuild x"
        self.assertEqual(hmm.command_line, "hmmbuild x")
        hmm.command_line = ""
        self.assertEqual(hmm.command_line, "")
        hmm.command_line = None
        self.assertIs(hmm.command_line, None)

    def test_name(self):
        abc = Alphabet.amino()
        hmm = HMM(abc, 100, b"test")

        self.assertEqual(hmm.name, b"test")
        hmm.name = b"Test"
        self.assertEqual(hmm.name, b"Test")
        hmm.name = b""
        self.assertEqual(hmm.name, b"")

    def test_accession(self):
        abc = Alphabet.amino()
        hmm = HMM(abc, 100, b"test")

        self.assertIs(hmm.accession, None)
        hmm.accession = b"TST001"
        self.assertEqual(hmm.accession, b"TST001")
        hmm.accession = b""
        self.assertEqual(hmm.accession, b"")
        hmm.accession = None
        self.assertIs(hmm.accession, None)

    def test_description(self):
        abc = Alphabet.amino()
        hmm = HMM(abc, 100, b"test")

        self.assertIs(hmm.description, None)
        hmm.description = b"A very important HMM"
        self.assertEqual(hmm.description, b"A very important HMM")
        hmm.description = b""
        self.assertEqual(hmm.description, b"")
        hmm.description = None
        self.assertIs(hmm.description, None)

    def test_consensus(self):
        # if not set, HMM is None
        abc = Alphabet.amino()
        hmm = HMM(abc, 100, b"test")
        self.assertIs(hmm.consensus, None)
        # if the HMM is fully configured, the consensus should be as
        # long as the number of nodes
        self.assertEqual(len(self.hmm.consensus), self.hmm.M)

    def test_set_consensus(self):
        abc = Alphabet.dna()
        hmm = HMM(abc, 100, b"test")
        self.assertIs(hmm.consensus, None)
        hmm.set_consensus()
        self.assertIsNot(hmm.consensus, None)
        self.assertEqual(len(hmm.consensus), hmm.M)

        seq = TextSequence(sequence="A"*hmm.M)
        hmm.set_consensus(seq.digitize(hmm.alphabet))
        self.assertEqual(hmm.consensus, seq.sequence)

    def test_set_consensus_error(self):
        dna = Alphabet.dna()
        prot = Alphabet.amino()
        hmm = HMM(dna, 100, b"test")
        with self.assertRaises(AlphabetMismatch):
            seq = TextSequence(sequence="Y"*hmm.M).digitize(prot)
            hmm.set_consensus(seq)
        with self.assertRaises(ValueError):
            seq = TextSequence(sequence="A"*(hmm.M - 1)).digitize(dna)
            hmm.set_consensus(seq)

    @unittest.skipIf(sys.implementation.name == "pypy", "`getsizeof` not supported on PyPY")
    def test_sizeof(self):
        dna = Alphabet.dna()
        hmm = HMM(dna, 100, b"test")
        self.assertGreater(sys.getsizeof(hmm), 0)

    @unittest.skipUnless(resource_files, "importlib.resources.files not available")
    def test_write(self):
        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "Thioesterase.hmm")
        with HMMFile(hmm_path) as hmm_file:
            hmm = next(hmm_file)

        buffer = io.BytesIO()
        hmm.write(buffer)

        self.assertNotEqual(buffer.tell(), 0)
        buffer.seek(0)

        with open(hmm_path, "rb") as f:
            # 1st line is skipped, cause it contains the format date, which will
            # obviously not be the same as the reference
            for line_written, line_ref in itertools.islice(zip(buffer, f), 1, None):
                self.assertEqual(line_written, line_ref)

    def test_write_empty(self):
        buffer = io.BytesIO()
        abc = Alphabet.amino()
        hmm = HMM(abc, 10, b"test")
        hmm.write(buffer)
        self.assertGreater(len(buffer.getvalue()), 0)

    @unittest.skipIf(platform.system() == "Darwin", "segfault at exit on MacOS")
    def test_write_error(self):
        buffer = io.StringIO()

        with self.assertRaises(EaselError) as ctx:
            self.hmm.write(buffer)
        self.assertEqual(ctx.exception.code, 27)
        self.assertIsInstance(ctx.exception.__cause__, TypeError)

    def test_renormalize(self):
        abc = Alphabet.dna()
        hmm = HMM(abc, 10, b"test")
        for i in range(1, hmm.M+1):
            for j in range(abc.K):
                hmm.match_emissions[i, j] = 25.0

        hmm.renormalize()
        for i in range(hmm.M+1):
            self.assertAlmostEqual(hmm.match_emissions[i].sum(), 1.0, places=5)

    def test_scale(self):
        abc = Alphabet.dna()
        hmm = HMM(abc, 10, b"test")
        for i in range(1, hmm.M+1):
            for j in range(abc.K):
                hmm.match_emissions[i, j] = 25.0

        hmm.scale(2.0)
        for i in range(1, hmm.M+1):
            for j in range(abc.K):
                self.assertAlmostEqual(hmm.match_emissions[i, j], 50.0, places=5)

    def test_zero(self):
        abc = Alphabet.dna()
        hmm = HMM(abc, 10, b"custom")
        for i in range(1, hmm.M+1):
            for j in range(abc.K):
                hmm.match_emissions[i, j] = 25.0
        self.assertEqual(hmm.name, b"custom")

        hmm.zero()
        for i in range(hmm.M+1):
            for j in range(abc.K):
                self.assertEqual(hmm.match_emissions[i, j], 0.0)
        self.assertEqual(hmm.name, b"custom")

    def test_pickle(self):
        h1 = self.hmm
        h2 = pickle.loads(pickle.dumps(h1))

        self.assertEqual(h1.alphabet, h2.alphabet)
        self.assertEqual(h1.M, h2.M)
        self.assertEqual(h1.transition_probabilities, h2.transition_probabilities)
        self.assertEqual(h1.insert_emissions, h2.insert_emissions)
        self.assertEqual(h1.match_emissions, h2.match_emissions)
        self.assertEqual(h1.name, h2.name)
        self.assertEqual(h1.accession, h2.accession)
        self.assertEqual(h1.description, h2.description)
        self.assertEqual(h1.checksum, h2.checksum)
        self.assertEqual(h1.composition, h2.composition)
        self.assertEqual(h1.consensus, h2.consensus)
        self.assertEqual(h1.consensus_structure, h2.consensus_structure)
        self.assertEqual(h1.consensus_accessibility, h2.consensus_accessibility)
        self.assertEqual(h1.reference, h2.reference)
        self.assertEqual(h1.model_mask, h2.model_mask)
        self.assertEqual(h1.command_line, h2.command_line)
        self.assertEqual(h1.nseq, h2.nseq)
        self.assertEqual(h1.nseq_effective, h2.nseq_effective)

        self.assertEqual(h1.creation_time, h2.creation_time)
        self.assertEqual(h1.cutoffs, h2.cutoffs)
        self.assertEqual(h1.evalue_parameters, h2.evalue_parameters)
        self.assertEqual(h1.composition, h2.composition)

        self.assertEqual(h1, h2)

    @unittest.skipUnless(resource_files, "importlib.resources.files not available")
    def test_no_cutoffs(self):
        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "Thioesterase.hmm")
        with HMMFile(hmm_path) as hmm_file:
            hmm = next(hmm_file)

        self.assertFalse(hmm.cutoffs.gathering_available())
        self.assertIs(hmm.cutoffs.gathering, None)
        self.assertIs(hmm.cutoffs.gathering1, None)
        self.assertIs(hmm.cutoffs.gathering2, None)
        self.assertFalse(hmm.cutoffs.noise_available())
        self.assertIs(hmm.cutoffs.noise, None)
        self.assertIs(hmm.cutoffs.noise1, None)
        self.assertIs(hmm.cutoffs.noise2, None)
        self.assertFalse(hmm.cutoffs.trusted_available())
        self.assertIs(hmm.cutoffs.trusted, None)
        self.assertIs(hmm.cutoffs.trusted1, None)
        self.assertIs(hmm.cutoffs.trusted2, None)

        hmm.cutoffs.gathering = (10.0, 12.0)
        self.assertTrue(hmm.cutoffs.gathering_available())
        self.assertFalse(hmm.cutoffs.noise_available())
        self.assertFalse(hmm.cutoffs.trusted_available())
        self.assertEqual(hmm.cutoffs.gathering, (10.0, 12.0))
        self.assertEqual(hmm.cutoffs.gathering1, 10.0)
        self.assertEqual(hmm.cutoffs.gathering2, 12.0)

        hmm.cutoffs.noise = (8.0, 5.0)
        self.assertTrue(hmm.cutoffs.gathering_available())
        self.assertTrue(hmm.cutoffs.noise_available())
        self.assertFalse(hmm.cutoffs.trusted_available())
        self.assertEqual(hmm.cutoffs.noise, (8.0, 5.0))
        self.assertEqual(hmm.cutoffs.noise1, 8.0)
        self.assertEqual(hmm.cutoffs.noise2, 5.0)

        hmm.cutoffs.trusted = (15.0, 14.0)
        self.assertTrue(hmm.cutoffs.gathering_available())
        self.assertTrue(hmm.cutoffs.noise_available())
        self.assertTrue(hmm.cutoffs.trusted_available())
        self.assertEqual(hmm.cutoffs.trusted, (15.0, 14.0))
        self.assertEqual(hmm.cutoffs.trusted1, 15.0)
        self.assertEqual(hmm.cutoffs.trusted2, 14.0)

    @unittest.skipUnless(resource_files, "importlib.resources.files not available")
    def test_cutoffs(self):
        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "PF02826.hmm")
        with HMMFile(hmm_path) as hmm_file:
            hmm = next(hmm_file)

        self.assertTrue(hmm.cutoffs.gathering_available())
        self.assertTrue(hmm.cutoffs.noise_available())
        self.assertTrue(hmm.cutoffs.trusted_available())

        hmm.cutoffs.gathering = None
        self.assertFalse(hmm.cutoffs.gathering_available())
        self.assertTrue(hmm.cutoffs.noise_available())
        self.assertTrue(hmm.cutoffs.trusted_available())
        self.assertIs(hmm.cutoffs.gathering, None)
        self.assertIs(hmm.cutoffs.gathering1, None)
        self.assertIs(hmm.cutoffs.gathering2, None)

        del hmm.cutoffs.noise
        self.assertFalse(hmm.cutoffs.gathering_available())
        self.assertFalse(hmm.cutoffs.noise_available())
        self.assertTrue(hmm.cutoffs.trusted_available())
        self.assertIs(hmm.cutoffs.noise, None)
        self.assertIs(hmm.cutoffs.noise1, None)
        self.assertIs(hmm.cutoffs.noise2, None)

        hmm.cutoffs.trusted = None
        self.assertFalse(hmm.cutoffs.gathering_available())
        self.assertFalse(hmm.cutoffs.noise_available())
        self.assertFalse(hmm.cutoffs.trusted_available())
        self.assertIs(hmm.cutoffs.trusted, None)
        self.assertIs(hmm.cutoffs.trusted1, None)
        self.assertIs(hmm.cutoffs.trusted2, None)

    @unittest.skipUnless(resource_files, "importlib.resources.files not available")
    def test_cutoffs_pickle(self):
        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "PF02826.hmm")
        with HMMFile(hmm_path) as hmm_file:
            hmm = next(hmm_file)
        hmm_pickled = pickle.loads(pickle.dumps(hmm))
        self.assertTrue(hmm_pickled.cutoffs.gathering_available())
        self.assertTrue(hmm_pickled.cutoffs.noise_available())
        self.assertTrue(hmm_pickled.cutoffs.trusted_available())
        self.assertEqual(hmm_pickled.cutoffs, hmm.cutoffs)



from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t, int64_t


cdef extern from "hmmer.h" nogil:

    cdef struct fm_data_s:
        uint64_t N
        uint32_t term_loc
        uint32_t seq_offset
        uint32_t ambig_offset
        uint32_t seq_cnt
        uint32_t ambig_cnt
        uint32_t overlap
        uint8_t* T
        uint8_t* BWT_mem
        uint8_t* BWT
        uint32_t* SA
        int64_t*  C
        uint32_t* occCnts_sb
        uint16_t* occCnts_b
    ctypedef fm_data_s FM_DATA

    ctypedef struct FM_CFG:
        pass



#include <pthread.h>

cdef extern from "pthread.h" nogil:

    ctypedef int pthread_mutex_t

    const pthread_mutex_t PTHREAD_MUTEX_INITIALIZER
    const pthread_mutex_t PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
    const pthread_mutex_t PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP

    int pthread_mutex_init(pthread_mutex_t *mutex, const void* mutexattr)
    int pthread_mutex_lock(pthread_mutex_t *mutex)
    int pthread_mutex_trylock(pthread_mutex_t *mutex)
    int pthread_mutex_unlock(pthread_mutex_t *mutex)
    int pthread_mutex_destroy(pthread_mutex_t *mutex)



import typing
import os
import contextlib

from ..easel import SSIWriter
from ..plan7 import HMM, Background, Profile

# --- hmmpress ---------------------------------------------------------------

def hmmpress(
    hmms: typing.Iterable[HMM],
    output: typing.Union[str, "os.PathLike[str]"],
) -> int:
    """Press several HMMs into a database.

    Calling this function will create 4 files at the given location:
    ``{output}.h3p`` (containing the optimized profiles),
    ``{output}.h3m`` (containing the binary HMMs),
    ``{output}.h3f`` (containing the MSV parameters), and
    ``{output}.h3i`` (the SSI index mapping the previous files).

    Arguments:
        hmms (iterable of `~pyhmmer.plan7.HMM`): The HMMs to be pressed
            together in the file.
        output (`str` or `os.PathLike`): The path to an output location
            where to write the different files.

    """
    DEFAULT_L = 400
    path = os.fspath(output)
    nmodel = 0

    with contextlib.ExitStack() as ctx:
        h3p = ctx.enter_context(open("{}.h3p".format(path), "wb"))
        h3m = ctx.enter_context(open("{}.h3m".format(path), "wb"))
        h3f = ctx.enter_context(open("{}.h3f".format(path), "wb"))
        h3i = ctx.enter_context(SSIWriter("{}.h3i".format(path)))
        fh = h3i.add_file(path, format=0)

        for hmm in hmms:
            # create the background model on the first iteration
            if nmodel == 0:
                bg = Background(hmm.alphabet)
                bg.L = DEFAULT_L

            # build the optimized models
            gm = Profile(hmm.M, hmm.alphabet)
            gm.configure(hmm, bg, DEFAULT_L)
            om = gm.to_optimized()

            # update the disk offsets of the optimized model to be written
            om.offsets.model = h3m.tell()
            om.offsets.profile = h3p.tell()
            om.offsets.filter = h3f.tell()

            # check that hmm has a name
            if hmm.name is None:
                raise ValueError("HMMs must have a name to be pressed.")
            # add the HMM name, and optionally the HMM accession to the index
            h3i.add_key(hmm.name, fh, om.offsets.model, 0, 0)
            if hmm.accession is not None:
                h3i.add_alias(hmm.accession, hmm.name)

            # write the HMM in binary format, and the optimized profile
            hmm.write(h3m, binary=True)
            om.write(h3f, h3p)
            nmodel += 1

    # return the number of written HMMs
    return nmodel






cdef extern from "esl_json.h" nogil:

    cdef enum esl_json_type_e:
        eslJSON_UNKNOWN = 0
        eslJSON_OBJECT  = 1
        eslJSON_ARRAY   = 2
        eslJSON_KEY     = 3
        eslJSON_STRING  = 4
        eslJSON_NUMBER  = 5
        eslJSON_BOOLEAN = 6
        eslJSON_NULL    = 7

    cdef enum esl_json_state_e:
        eslJSON_OBJ_NONE      = 0
        eslJSON_OBJ_OPEN      = 1
        eslJSON_OBJ_COLON     = 2
        eslJSON_OBJ_COMMA     = 3
        eslJSON_ARR_OPEN      = 4
        eslJSON_ARR_COMMA     = 5
        eslJSON_STR_OPEN      = 6
        eslJSON_STR_CHAR      = 7
        eslJSON_STR_BACKSLASH = 8
        eslJSON_STR_PROTECTED = 9
        eslJSON_STR_UNICODE   = 10
        eslJSON_KEY_OPEN      = 11
        eslJSON_KEY_CHAR      = 12
        eslJSON_KEY_BACKSLASH = 13
        eslJSON_KEY_PROTECTED = 14
        eslJSON_KEY_UNICODE   = 15
        eslJSON_NUM_SIGN      = 16
        eslJSON_NUM_ZERO      = 17
        eslJSON_NUM_NONZERO   = 18
        eslJSON_NUM_LEADDIGIT = 19
        eslJSON_NUM_POINT     = 20
        eslJSON_NUM_FRACDIGIT = 21
        eslJSON_NUM_EXP       = 22
        eslJSON_NUM_EXPSIGN   = 23
        eslJSON_NUM_EXPDIGIT  = 24
        eslJSON_VAL_TRUE      = 25
        eslJSON_VAL_FALSE     = 26
        eslJSON_VAL_NULL      = 27
        eslJSON_VAL_INOBJ     = 28
        eslJSON_VAL_INARR     = 29
        eslJSON_STR_ASKEY     = 30

    ctypedef struct ESL_JSON_TOK:
        esl_json_type_e type
        esl_pos_t startpos
        esl_pos_t endpos
        int nchild
        int firstchild
        int lastchild
        int nextsib
        int linenum
        int linepos

    ctypedef struct ESL_JSON:
        ESL_JSON_TOK* tok
        int ntok
        int nalloc
        int redline

    ctypedef struct ESL_JSON_PARSER:
        esl_json_state_e state
        ESL_STACK* pda
        int curridx
        int codelen
        esl_pos_t pos
        int linenum
        int linepos

    # Full and incremental JSON parsing
    int esl_json_Parse(ESL_BUFFER *bf, ESL_JSON **ret_pi);
    int esl_json_PartialParse(ESL_JSON_PARSER *parser, ESL_JSON *pi, const char *s, esl_pos_t n, esl_pos_t *ret_nused, char *errbuf);

    # ESL_JSON
    ESL_JSON *esl_json_Create   (void);
    int       esl_json_Grow     (ESL_JSON *pi);
    size_t    esl_json_Sizeof   (ESL_JSON *pi);
    size_t    esl_json_MinSizeof(ESL_JSON *pi);
    int       esl_json_Reuse    (ESL_JSON *pi);
    void      esl_json_Destroy  (ESL_JSON *pi);

    # ESL_JSON_PARSER
    ESL_JSON_PARSER *esl_json_parser_Create(void);
    void             esl_json_parser_Destroy(ESL_JSON_PARSER *parser);

    # Accessing tokenized data
    char      *esl_json_GetMem   (const ESL_JSON *pi, int idx, const ESL_BUFFER *bf);
    esl_pos_t  esl_json_GetLen   (const ESL_JSON *pi, int idx, const ESL_BUFFER *bf);
    int        esl_json_ReadInt  (const ESL_JSON *pi, int idx,       ESL_BUFFER *bf, int   *ret_i);
    int        esl_json_ReadFloat(const ESL_JSON *pi, int idx,       ESL_BUFFER *bf, float *ret_x);

    # Debugging, development 
    int   esl_json_Validate(const ESL_JSON *pi, const ESL_BUFFER *bf, char *errbuf);
    char *esl_json_DecodeType(enum esl_json_type_e type);
    int   esl_json_Dump(FILE *fp, ESL_JSON *pi);
    int   esl_json_SampleDirty(ESL_RANDOMNESS *rng, char **ret_s, int *ret_n);



import copy
import io
import itertools
import os
import shutil
import unittest
import tempfile
import pickle

import pyhmmer
from pyhmmer.errors import EaselError, AlphabetMismatch
from pyhmmer.easel import (
    Alphabet, 
    GeneticCode,
    TextSequence, 
    DigitalSequence, 
    DigitalSequenceBlock, 
    VectorU8
)

class TestGeneticCode(unittest.TestCase):

    def test_invalid_table(self):
        with self.assertRaises(ValueError):
            gencode = GeneticCode(42)

    def test_translate_empty(self):
        gencode = GeneticCode()
        prot = gencode.translate(bytearray())
        self.assertEqual(len(prot), 0)
        self.assertEqual(prot, VectorU8())

    def test_length_error(self):
        gencode = GeneticCode()
        with self.assertRaises(ValueError):
            gencode.translate(bytearray([0, 0, 0, 1, 1]))

    def test_pickle(self):
        gencode = GeneticCode(11)
        gencode2 = pickle.loads(pickle.dumps(gencode))
        self.assertEqual(gencode.translation_table, gencode2.translation_table)
        self.assertEqual(gencode.nucleotide_alphabet, gencode2.nucleotide_alphabet)
        self.assertEqual(gencode.amino_alphabet, gencode2.amino_alphabet)


from libc.stdio cimport FILE

from libeasel.fileparser cimport ESL_FILEPARSER
from libeasel.random cimport ESL_RANDOMNESS


cdef extern from "esl_mixdchlet.h" nogil:

    ctypedef struct ESL_MIXDCHLET:
        double* q
        double** alpha
        int Q
        int K
        double* postq

    ESL_MIXDCHLET *esl_mixdchlet_Create(int Q, int K)
    void           esl_mixdchlet_Destroy(ESL_MIXDCHLET *dchl)

    double         esl_mixdchlet_logp_c      (ESL_MIXDCHLET *dchl, double *c)
    int            esl_mixdchlet_MPParameters(ESL_MIXDCHLET *dchl, double *c, double *p)

    int            esl_mixdchlet_Fit(double **c, int N, ESL_MIXDCHLET *dchl, double *opt_nll)
    int            esl_mixdchlet_Sample(ESL_RANDOMNESS *rng, ESL_MIXDCHLET *dchl)

    int            esl_mixdchlet_Read(ESL_FILEPARSER *efp, ESL_MIXDCHLET **ret_dchl)
    int            esl_mixdchlet_Write    (FILE *fp, const ESL_MIXDCHLET *dchl)
    int            esl_mixdchlet_WriteJSON(FILE *fp, const ESL_MIXDCHLET *dchl)

    int            esl_mixdchlet_Validate(const ESL_MIXDCHLET *dchl, char *errmsg)
    int            esl_mixdchlet_Compare(const ESL_MIXDCHLET *d1, const ESL_MIXDCHLET *d2, double tol)
    int            esl_mixdchlet_Dump(FILE *fp, const ESL_MIXDCHLET *dchl)



from libc.stdint cimport int64_t, uint32_t
from posix.types cimport off_t

from libeasel cimport ESL_DSQ, esl_pos_t
from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.bitfield cimport ESL_BITFIELD
from libeasel.keyhash cimport ESL_KEYHASH
from libeasel.random cimport ESL_RANDOMNESS


cdef extern from "esl_msa.h" nogil:

    cdef enum:
        eslMSA_TC1 = 0
        eslMSA_TC2 = 1
        eslMSA_GA1 = 2
        eslMSA_GA2 = 3
        eslMSA_NC1 = 4
        eslMSA_NC2 = 5
        eslMSA_NCUTS = 6

    cdef enum:
        eslMSA_HASWGTS = 0b01
        eslMSA_DIGITAL = 0b10

    ctypedef struct ESL_MSA:
        char** aseq
        char** sqname
        double* wgt
        int64_t alen
        int nseq
        int flags

        ESL_ALPHABET* abc
        ESL_DSQ** ax

        char* name
        char* desc
        char* acc
        char* au
        char* ss_cons
        char* sa_cons
        char* pp_cons
        char* rf
        char* mm
        char** sqacc
        char** sqdesc
        char** ss
        char** sa
        char** pp
        float[eslMSA_NCUTS] cutoff
        float[eslMSA_NCUTS] cutset

        int sqalloc
        int64_t* sqlen
        int64_t* sslen
        int64_t* salen
        int64_t* pplen
        int lastidx

        char** comment
        int ncomment
        int alloc_ncomment

        char** gf_tag
        char** gf
        int ngf
        int alloc_ngf

        char*** gs_tag
        char*** gs
        int ngs

        char** gr_tag
        char*** gr
        int ngr

        ESL_KEYHASH* index
        ESL_KEYHASH* gs_idx
        ESL_KEYHASH* gc_idx
        ESL_KEYHASH* gr_idx

        off_t offset


    # The ESL_MSA object
    ESL_MSA *esl_msa_Create(int nseq, int64_t alen)
    int      esl_msa_Expand(ESL_MSA *msa)
    int      esl_msa_Copy(const ESL_MSA *msa, ESL_MSA *new_msa)
    ESL_MSA *esl_msa_Clone(const ESL_MSA *msa)
    void     esl_msa_Destroy(ESL_MSA *msa)

    # Digital mode MSA's
    int      esl_msa_GuessAlphabet(const ESL_MSA *msa, int *ret_type)
    ESL_MSA *esl_msa_CreateDigital(const ESL_ALPHABET *abc, int nseq, int64_t alen)
    int      esl_msa_Digitize(const ESL_ALPHABET *abc, ESL_MSA *msa, char *errmsg)
    int      esl_msa_Textize(ESL_MSA *msa)
    int      esl_msa_ConvertDegen2X(ESL_MSA *msa)

    # Setting or checking data fields in an ESL_MSA
    int esl_msa_SetName          (ESL_MSA *msa, const char *s, esl_pos_t n)
    int esl_msa_SetDesc          (ESL_MSA *msa, const char *s, esl_pos_t n)
    int esl_msa_SetAccession     (ESL_MSA *msa, const char *s, esl_pos_t n)
    int esl_msa_SetAuthor        (ESL_MSA *msa, const char *s, esl_pos_t n)
    int esl_msa_SetSeqName       (ESL_MSA *msa, int idx, const char *s, esl_pos_t n)
    int esl_msa_SetSeqAccession  (ESL_MSA *msa, int idx, const char *s, esl_pos_t n)
    int esl_msa_SetSeqDescription(ESL_MSA *msa, int idx, const char *s, esl_pos_t n)
    int esl_msa_SetDefaultWeights(ESL_MSA *msa)

    int esl_msa_FormatName          (ESL_MSA *msa, const char *name,    ...)
    int esl_msa_FormatDesc          (ESL_MSA *msa, const char *desc,    ...)
    int esl_msa_FormatAccession     (ESL_MSA *msa, const char *acc,     ...)
    int esl_msa_FormatAuthor        (ESL_MSA *msa, const char *author,  ...)
    int esl_msa_FormatSeqName       (ESL_MSA *msa, int idx, const char *name, ...)
    int esl_msa_FormatSeqAccession  (ESL_MSA *msa, int idx, const char *acc, ...)
    int esl_msa_FormatSeqDescription(ESL_MSA *msa, int idx, const char *desc, ...)

    int esl_msa_AddComment(ESL_MSA *msa, char *p,   esl_pos_t n)
    int esl_msa_AddGF     (ESL_MSA *msa, char *tag, esl_pos_t taglen,            char *value, esl_pos_t vlen)
    int esl_msa_AddGS     (ESL_MSA *msa, char *tag, esl_pos_t taglen, int sqidx, char *value, esl_pos_t vlen)
    int esl_msa_AppendGC  (ESL_MSA *msa, char *tag, char *value)
    int esl_msa_AppendGR  (ESL_MSA *msa, char *tag, int sqidx, char *value)

    int esl_msa_CheckUniqueNames(const ESL_MSA *msa)

    # Miscellaneous functions for manipulating MSAs
    int esl_msa_ReasonableRF(ESL_MSA *msa, double symfrac, int useconsseq, char *rfline)
    int esl_msa_MarkFragments (const ESL_MSA *msa, float fragthresh, ESL_BITFIELD **ret_fragassign)
    int esl_msa_MarkFragments_old(ESL_MSA *msa, double fragthresh)
    int esl_msa_SequenceSubset(const ESL_MSA *msa, const int *useme, ESL_MSA **ret_new)
    int esl_msa_ColumnSubset (ESL_MSA *msa, char *errbuf, const int *useme)
    int esl_msa_MinimGaps    (ESL_MSA *msa, char *errbuf, const char *gaps, int consider_rf)
    int esl_msa_MinimGapsText(ESL_MSA *msa, char *errbuf, const char *gaps, int consider_rf, int fix_bps)
    int esl_msa_NoGaps       (ESL_MSA *msa, char *errbuf, const char *gaps)
    int esl_msa_NoGapsText   (ESL_MSA *msa, char *errbuf, const char *gaps, int fix_bps)
    int esl_msa_SymConvert(ESL_MSA *msa, const char *oldsyms, const char *newsyms)
    int esl_msa_Checksum(const ESL_MSA *msa, uint32_t *ret_checksum)

    int esl_msa_RemoveBrokenBasepairsFromSS(char *ss, char *errbuf, int len, const int *useme)
    int esl_msa_RemoveBrokenBasepairs(ESL_MSA *msa, char *errbuf, const int *useme)

    int esl_msa_ReverseComplement(ESL_MSA *msa)
    int esl_msa_Hash(ESL_MSA *msa)
    int esl_msa_FlushLeftInserts(ESL_MSA *msa)

    # Debugging, testing, development
    int      esl_msa_Validate(const ESL_MSA *msa, char *errmsg)
    ESL_MSA *esl_msa_CreateFromString(const char *s, int fmt)
    int      esl_msa_Compare         (ESL_MSA *a1, ESL_MSA *a2)
    int      esl_msa_CompareMandatory(ESL_MSA *a1, ESL_MSA *a2)
    int      esl_msa_CompareOptional (ESL_MSA *a1, ESL_MSA *a2)
    int      esl_msa_Sample(ESL_RANDOMNESS *rng, ESL_ALPHABET *abc, int max_nseq, int max_alen, ESL_MSA **ret_msa)



import abc
import collections
import math
import io
import itertools
import os
import platform
import unittest
import tempfile
import threading
import multiprocessing.resource_sharer

import pyhmmer
from pyhmmer.plan7 import HMM, Pipeline, HMMFile, HMMPressedFile, TopHits, Hit
from pyhmmer.easel import Alphabet, DigitalMSA, MSAFile, SequenceFile, TextSequence

from .utils import resource_files

class _TestSearch(metaclass=abc.ABCMeta):

    def tearDown(self):
        multiprocessing.resource_sharer.stop()
        self.assertEqual(threading.active_count(), 1, threading.enumerate())

    @abc.abstractmethod
    def get_hits(self, hmm, sequences):
        return NotImplemented

    def get_hits_multi(self, hmms, sequences):
        return [self.get_hits(hmm, sequences) for hmm in hmms]

    @staticmethod
    def table(name):
        path = resource_files(__package__).joinpath("data", "tables", name)
        return path.open()

    @staticmethod
    def hmm_file(name):
        path = resource_files(__package__).joinpath("data", "hmms", "txt", "{}.hmm".format(name))
        return HMMFile(path)

    @staticmethod
    def seqs_file(name, digital=False):
        seqs_path = resource_files(__package__).joinpath("data", "seqs", "{}.faa".format(name))
        return SequenceFile(seqs_path, digital=digital)

    @unittest.skipUnless(resource_files, "importlib.resources not available")
    def test_thioestherase(self):
        # $ hmmsearch data/hmm/Thioesterase.hmm data/seqs/938293.PRJEB85.HG003687.faa
        #
        # Query:       Thioesterase  [M=243]
        # Scores for complete sequences (score includes all domains):
        #    --- full sequence ---   --- best 1 domain ---    -#dom-
        #     E-value  score  bias    E-value  score  bias    exp  N  Sequence                    Description
        #     ------- ------ -----    ------- ------ -----   ---- --  --------                    -----------
        #   ------ inclusion threshold ------
        #       0.096    8.6   1.5       0.14    8.1   1.5    1.2  1  938293.PRJEB85.HG003687_113  # 108502 # 109284 # 1 # ID=8_113;partial=00;start_type=ATG;rbs_motif=GGAG/GAGG;rbs_spacer=5-10bp;gc_cont=0.341
        #
        # Domain annotation for each sequence (and alignments):
        # >> 938293.PRJEB85.HG003687_113  # 108502 # 109284 # 1 # ID=8_113;partial=00;start_type=ATG;rbs_motif=GGAG/GAGG;rbs_spacer=5-10bp;gc_cont=0.341
        #    #    score  bias  c-Evalue  i-Evalue hmmfrom  hmm to    alifrom  ali to    envfrom  env to     acc
        #  ---   ------ ----- --------- --------- ------- -------    ------- -------    ------- -------    ----
        #    1 ?    8.1   1.5   6.5e-05      0.14      79      93 ..     115     129 ..     115     129 .. 0.96
        #
        #   Alignments for each domain:
        #   == domain 1  score: 8.1 bits;  conditional E-value: 6.5e-05
        #                  Thioesterase  79 GWSfGGvlAyEmArq 93
        #                                   G+S+GG +A ++A++
        #   938293.PRJEB85.HG003687_113 115 GHSMGGSVAVAIAHE 129
        #                                   9************96 PP
        with self.hmm_file("Thioesterase") as hmm_file:
            hmm = hmm_file.read()
        with self.seqs_file("938293.PRJEB85.HG003687", digital=True) as seqs_file:
            seqs = seqs_file.read_block()

        hits = self.get_hits(hmm, seqs)
        self.assertEqual(len(hits), 1)

        hits.sort()

        hit = hits[0]
        self.assertEqual(hit.name, b"938293.PRJEB85.HG003687_113")
        self.assertAlmostEqual(hit.score, 8.6, delta=0.1)  # printed with %6.1f
        self.assertAlmostEqual(hit.bias, 1.5, delta=0.1)  # printed with  %5.1f
        self.assertAlmostEqual(hit.evalue, 0.096, delta=0.01)  # printed with %9.2g
        self.assertEqual(len(hit.domains), 1)

        domain = hit.domains[0]
        self.assertAlmostEqual(domain.score, 8.1, delta=0.1)
        self.assertAlmostEqual(domain.bias, 1.5, delta=0.1)
        self.assertAlmostEqual(domain.i_evalue, 0.14, places=2)  # printed with %9.2g
        self.assertAlmostEqual(domain.c_evalue, 6.5e-05, places=2)  # printed with %9.2g
        self.assertEqual(domain.alignment.target_from, 115)
        self.assertEqual(domain.alignment.target_to, 129)
        self.assertEqual(domain.alignment.target_length, 261)
        self.assertEqual(domain.alignment.hmm_from, 79)
        self.assertEqual(domain.alignment.hmm_to, 93)
        self.assertEqual(domain.alignment.hmm_length, 243)
        self.assertEqual(domain.env_from, 115)
        self.assertEqual(domain.env_to, 129)

        last = hit.domains[-1]
        self.assertEqual(domain.alignment.hmm_name, last.alignment.hmm_name)

    @unittest.skipUnless(resource_files, "importlib.resources not available")
    def test_pf02826_file(self):
        with self.hmm_file("PF02826") as hmm_file:
            hmm = hmm_file.read()
        with self.seqs_file("938293.PRJEB85.HG003687", digital=True) as seqs_file:
            hits = self.get_hits(hmm, seqs_file)
            self.assertEqual(len(hits), 22)

        hits.sort()

        with self.table("PF02826.tbl") as table:
            lines = filter(lambda line: not line.startswith("#"), table)
            for line, hit in itertools.zip_longest(lines, hits):
                fields = list(filter(None, line.strip().split(" ")))
                self.assertIsNot(line, None)
                self.assertIsNot(hit, None)
                self.assertEqual(hit.name.decode(), fields[0])
                if fields[1] == "-":
                    self.assertIs(hit.accession, None)
                else:
                    self.assertEqual(hit.accession.decode(), fields[1])
                self.assertAlmostEqual(hit.score, float(fields[5]), delta=0.1)
                self.assertAlmostEqual(hit.bias, float(fields[6]), delta=0.1)
                self.assertAlmostEqual(hit.evalue, float(fields[4]), delta=0.1)

    @unittest.skipUnless(resource_files, "importlib.resources not available")
    def test_pf02826_block(self):
        with self.hmm_file("PF02826") as hmm_file:
            hmm = hmm_file.read()
        with self.seqs_file("938293.PRJEB85.HG003687", digital=True) as seqs_file:
            seqs = seqs_file.read_block()

        hits = self.get_hits(hmm, seqs)
        self.assertEqual(len(hits), 22)

        hits.sort()

        with self.table("PF02826.tbl") as table:
            lines = filter(lambda line: not line.startswith("#"), table)
            for line, hit in itertools.zip_longest(lines, hits):
                fields = list(filter(None, line.strip().split(" ")))
                self.assertIsNot(line, None)
                self.assertIsNot(hit, None)
                self.assertEqual(hit.name.decode(), fields[0])
                if fields[1] == "-":
                    self.assertIs(hit.accession, None)
                else:
                    self.assertEqual(hit.accession.decode(), fields[1])
                self.assertAlmostEqual(hit.score, float(fields[5]), delta=0.1)
                self.assertAlmostEqual(hit.bias, float(fields[6]), delta=0.1)
                self.assertAlmostEqual(hit.evalue, float(fields[4]), delta=0.1)

    @unittest.skipUnless(resource_files, "importlib.resources not available")
    def test_rrefam(self):
        with self.seqs_file("938293.PRJEB85.HG003687", digital=True) as seqs_file:
            seqs = seqs_file.read_block()

        with self.hmm_file("RREFam") as hmm_file:
            all_hits = self.get_hits_multi(list(hmm_file), seqs)

        hits = (hit for hits in all_hits for hit in hits)
        with self.table("RREFam.tbl") as table:
            lines = filter(lambda line: not line.startswith("#"), table)
            for line, hit in itertools.zip_longest(lines, hits):
                fields = list(filter(None, line.strip().split(" ")))
                self.assertIsNot(line, None)
                self.assertIsNot(hit, None)
                self.assertEqual(hit.name.decode(), fields[0])
                if fields[1] == "-":
                    self.assertIs(hit.accession, None)
                else:
                    self.assertEqual(hit.accession.decode(), fields[1])
                self.assertAlmostEqual(hit.score, float(fields[5]), delta=0.1)
                self.assertAlmostEqual(hit.bias, float(fields[6]), delta=0.1)
                self.assertAlmostEqual(hit.evalue, float(fields[4]), delta=0.1)

        domains = [
            domain for hits in all_hits for hit in hits for domain in hit.domains
        ]
        with self.table("RREFam.domtbl") as table:
            lines = filter(lambda line: not line.startswith("#"), table)
            for line, domain in itertools.zip_longest(lines, domains):
                fields = list(filter(None, line.strip().split(" ")))
                self.assertIsNot(line, None)
                self.assertIsNot(domain, None)
                self.assertEqual(domain.hit.hits.Z, len(seqs))
                self.assertEqual(domain.hit.name.decode(), fields[0])
                self.assertAlmostEqual(domain.score, float(fields[13]), places=1)
                self.assertAlmostEqual(domain.bias, float(fields[14]), places=1)
                self.assertEqual(f"{domain.c_evalue:9.2g}", f"{float(fields[11]):9.2g}")
                self.assertEqual(f"{domain.i_evalue:9.2g}", f"{float(fields[12]):9.2g}")

    @unittest.skipUnless(resource_files, "importlib.resources not available")
    def test_hmm_vs_optimized_profile(self):
        with self.hmm_file("PF02826") as hmm_file:
            hmm = next(hmm_file)
            bg = pyhmmer.plan7.Background(hmm.alphabet)
            profile = pyhmmer.plan7.Profile(hmm.M, hmm.alphabet)
            profile.configure(hmm, bg, 100)
        with self.seqs_file("938293.PRJEB85.HG003687", digital=True) as seqs_file:
            seqs = seqs_file.read_block()

        hits_hmm = self.get_hits(hmm, seqs)
        self.assertEqual(len(hits_hmm), 22)

        hits_oprofile = self.get_hits(profile.to_optimized(), seqs)
        self.assertEqual(len(hits_oprofile), 22)

        for hit_hmm, hit_oprofile in itertools.zip_longest(hits_hmm, hits_oprofile):
            self.assertEqual(hit_hmm.name, hit_oprofile.name)
            self.assertEqual(hit_hmm.accession, hit_oprofile.accession)
            self.assertEqual(hit_hmm.description, hit_oprofile.description)
            self.assertEqual(hit_hmm.score, hit_oprofile.score)
            self.assertEqual(hit_hmm.pre_score, hit_oprofile.pre_score)
            self.assertEqual(hit_hmm.sum_score, hit_oprofile.sum_score)
            self.assertEqual(hit_hmm.bias, hit_oprofile.bias)
            self.assertEqual(hit_hmm.evalue, hit_oprofile.evalue)
            self.assertEqual(hit_hmm.pvalue, hit_oprofile.pvalue)
            for dom_hmm, dom_oprofile in itertools.zip_longest(
                hit_hmm.domains, hit_oprofile.domains
            ):
                self.assertEqual(dom_hmm.env_from, dom_oprofile.env_from)
                self.assertEqual(dom_hmm.env_to, dom_oprofile.env_to)
                self.assertEqual(dom_hmm.score, dom_oprofile.score)
                self.assertEqual(dom_hmm.bias, dom_oprofile.bias)
                self.assertEqual(dom_hmm.correction, dom_oprofile.correction)
                self.assertEqual(dom_hmm.envelope_score, dom_oprofile.envelope_score)
                self.assertEqual(dom_hmm.c_evalue, dom_oprofile.c_evalue)
                self.assertEqual(dom_hmm.i_evalue, dom_oprofile.i_evalue)
                self.assertEqual(dom_hmm.pvalue, dom_oprofile.pvalue)


class TestHmmsearch(_TestSearch, unittest.TestCase):
    parallel = "queries"

    @staticmethod
    def _random_sequences(n=20):
        rng = pyhmmer.easel.Randomness(42)
        alphabet = Alphabet.amino()
        return pyhmmer.easel.DigitalSequenceBlock(
            alphabet,
            [
                pyhmmer.easel.DigitalSequence.sample(alphabet, 200, rng)
                for _ in range(10)
            ]
        )

    def get_hits(self, hmm, seqs):
        return list(pyhmmer.hmmsearch(hmm, seqs, parallel=self.parallel))[0]

    def get_hits_multi(self, hmms, seqs):
        return list(pyhmmer.hmmsearch(hmms, seqs, parallel=self.parallel))

    def test_callback_error(self):

        class MyException(Exception):
            pass

        def callback(hmm, total):
            raise MyException("oopsie")

        rng = pyhmmer.easel.Randomness(42)
        alphabet = Alphabet.amino()
        hmm = HMM.sample(alphabet, 100, rng)
        seqs = self._random_sequences()

        hits = pyhmmer.hmmsearch(hmm, seqs, cpus=1, callback=callback, parallel=self.parallel)
        with self.assertRaises(MyException):
            hit = next(hits)

        hits = pyhmmer.hmmsearch(hmm, seqs, cpus=2, callback=callback, parallel=self.parallel)
        with self.assertRaises(MyException):
            hit = next(hits)

    def test_background_error(self):
        # check that errors occuring in worker threads are recovered and raised
        # in the main threads (a common error is mismatching the HMM and the
        # sequence alphabets).
        rng = pyhmmer.easel.Randomness(42)
        seqs = [TextSequence().digitize(Alphabet.dna())]
        hmm = HMM.sample(Alphabet.amino(), 100, rng)
        self.assertRaises(ValueError, self.get_hits, hmm, seqs)

    def test_no_queries(self):
        seqs = self._random_sequences()
        hits = pyhmmer.hmmsearch([], seqs, parallel=self.parallel)
        self.assertIs(None, next(hits, None))


class TestHmmsearchSingle(TestHmmsearch, unittest.TestCase):
    def get_hits(self, hmm, seqs):
        return list(pyhmmer.hmmsearch(hmm, seqs, cpus=1, parallel=self.parallel))[0]

    def get_hits_multi(self, hmm, seqs):
        return list(pyhmmer.hmmsearch(hmm, seqs, cpus=1, parallel=self.parallel))

    def test_no_queries(self):
        seqs = self._random_sequences()
        hits = pyhmmer.hmmsearch([], seqs, cpus=1, parallel=self.parallel)
        self.assertIs(None, next(hits, None))

@unittest.skipIf(platform.system() == "Darwin", "may deadlock on MacOS")
class TestHmmsearchProcess(TestHmmsearch, unittest.TestCase):
    def get_hits(self, hmm, seqs):
        return list(pyhmmer.hmmsearch(hmm, seqs, cpus=2, backend="multiprocessing", parallel=self.parallel))[0]

    def get_hits_multi(self, hmm, seqs):
        return list(pyhmmer.hmmsearch(hmm, seqs, cpus=2, backend="multiprocessing", parallel=self.parallel))

    def test_no_queries(self):
        seqs = self._random_sequences()
        hits = pyhmmer.hmmsearch([], seqs, cpus=2, backend="multiprocessing", parallel=self.parallel)
        self.assertIs(None, next(hits, None))


class TestHmmsearchReverse(TestHmmsearch):
    parallel = "targets"

    @unittest.skip("unsupported parallel mode with sequence file")
    def test_pf02826_file(self):
        pass


class TestHmmsearchReverseSingle(TestHmmsearchSingle):
    parallel = "targets"

    @unittest.skip("unsupported parallel mode with sequence file")
    def test_pf02826_file(self):
        pass


@unittest.skipIf(platform.system() == "Darwin", "may deadlock on MacOS")
class TestHmmsearchReverseProcess(TestHmmsearchProcess):
    parallel = "targets"

    @unittest.skip("unsupported parallel mode with sequence file")
    def test_pf02826_file(self):
        pass


class TestPipelinesearch(_TestSearch, unittest.TestCase):
    def get_hits(self, hmm, seqs):
        pipeline = Pipeline(alphabet=hmm.alphabet)
        hits = pipeline.search_hmm(hmm, seqs)
        return hits


class TestHmmpress(unittest.TestCase):
    def setUp(self):
        self.tmp = tempfile.NamedTemporaryFile(suffix=".hmm", delete=False).name

    def tearDown(self):
        for ext in ["", ".h3m", ".h3p", ".h3i", ".h3f"]:
            if os.path.exists(self.tmp + ext):
                os.remove(self.tmp + ext)

    @unittest.skipUnless(resource_files, "importlib.resources not available")
    def test_roundtrip(self):
        db_folder = resource_files(__package__).joinpath("data", "hmms", "db")
        self.hmm = db_folder.joinpath("Thioesterase.hmm")
        self.h3p = db_folder.joinpath("Thioesterase.hmm.h3p")
        self.h3m = db_folder.joinpath("Thioesterase.hmm.h3m")
        self.h3f = db_folder.joinpath("Thioesterase.hmm.h3f")
        self.h3i = db_folder.joinpath("Thioesterase.hmm.h3f")
        with HMMFile(self.hmm) as hmms:
            n = pyhmmer.hmmer.hmmpress(hmms, self.tmp)
            self.assertEqual(n, 1)
        with HMMFile(self.tmp) as hmm_file:
            hmm = next(hmm_file)
            self.assertEqual(hmm.name, b"Thioesterase")


class TestPhmmer(unittest.TestCase):

    @staticmethod
    def _random_sequences(n=20):
        rng = pyhmmer.easel.Randomness(42)
        alphabet = Alphabet.amino()
        return pyhmmer.easel.DigitalSequenceBlock(
            alphabet,
            [
                pyhmmer.easel.DigitalSequence.sample(alphabet, 200, rng)
                for _ in range(10)
            ]
        )

    @staticmethod
    def table(name):
        path = resource_files(__package__).joinpath("data", "tables", name)
        return path.open()

    @unittest.skipUnless(resource_files, "importlib.resources not available")
    def test_callback_error(self):

        class MyException(Exception):
            pass

        def callback(hmm, total):
            raise MyException("oopsie")

        alphabet = Alphabet.amino()
        path = resource_files(__package__).joinpath("data", "seqs", "PKSI.faa")
        with SequenceFile(path, digital=True, alphabet=alphabet) as seqs_file:
            seqs = seqs_file.read_block()

        hits = pyhmmer.phmmer(seqs[-1:], seqs, cpus=1, callback=callback)
        with self.assertRaises(MyException):
            hit = next(hits)

        hits = pyhmmer.phmmer(seqs[-1:], seqs, cpus=2, callback=callback)
        with self.assertRaises(MyException):
            hit = next(hits)

    def test_no_queries(self):
        seqs = self._random_sequences()
        hits = pyhmmer.phmmer([], seqs, cpus=1)
        self.assertIs(None, next(hits, None))

    @unittest.skipUnless(resource_files, "importlib.resources not available")
    def test_pksi(self):
        alphabet = Alphabet.amino()
        path = resource_files(__package__).joinpath("data", "seqs", "PKSI.faa")
        with SequenceFile(path, digital=True, alphabet=alphabet) as seqs_file:
            seqs = seqs_file.read_block()
        hits = next(pyhmmer.phmmer(seqs[-1:], seqs, cpus=1))
        hits.sort()

        with self.table("A0A089QRB9.domtbl") as table:
            lines = iter(filter(lambda line: not line.startswith("#"), table))
            it = ((hit, domain) for hit in hits for domain in hit.domains)
            for line, (hit, domain) in itertools.zip_longest(lines, it):
                self.assertIsNot(line, None)
                self.assertIsNot(hit, None)
                fields = list(filter(None, line.strip().split(" ")))

                self.assertEqual(hit.name.decode(), fields[0])
                self.assertAlmostEqual(hit.score, float(fields[7]), delta=0.1)
                self.assertAlmostEqual(hit.bias, float(fields[8]), delta=0.1)
                self.assertAlmostEqual(hit.evalue, float(fields[6]), delta=0.1)

                self.assertAlmostEqual(domain.i_evalue, float(fields[12]), delta=0.1)
                self.assertAlmostEqual(domain.score, float(fields[13]), delta=0.1)

                self.assertEqual(domain.alignment.hmm_from, int(fields[15]))
                self.assertEqual(domain.alignment.hmm_to, int(fields[16]))
                self.assertEqual(domain.alignment.target_from, int(fields[17]))
                self.assertEqual(domain.alignment.target_to, int(fields[18]))
                self.assertEqual(domain.env_from, int(fields[19]))
                self.assertEqual(domain.env_to, int(fields[20]))


@unittest.skipUnless(resource_files, "importlib.resources not available")
class TestJackhmmer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with cls.seqs_file("PKSI", digital=True) as seqs_file:
            cls.pksi = seqs_file.read_block()

    @staticmethod
    def table(name):
        path = resource_files(__package__).joinpath("data", "tables", name)
        return path.open()

    @staticmethod
    def seqs_file(name, digital=False):
        seqs_path = resource_files(__package__).joinpath("data", "seqs", "{}.faa".format(name))
        return SequenceFile(seqs_path, digital=digital)

    @staticmethod
    def hmm_file(name):
        path = resource_files(__package__).joinpath("data", "hmms", "txt", "{}.hmm".format(name))
        return HMMFile(path)

    def test_callback_error(self):

        class MyException(Exception):
            pass

        def callback(hmm, total):
            raise MyException("oopsie")

        seqs = self.pksi
        results = pyhmmer.jackhmmer(seqs[-1:], seqs, cpus=1, max_iterations=1, callback=callback)
        with self.assertRaises(MyException):
            result = next(results)

        results = pyhmmer.jackhmmer(seqs[-1:], seqs, cpus=2, max_iterations=1, callback=callback)
        with self.assertRaises(MyException):
            result = next(results)

    def test_no_queries(self):
        seqs = self.pksi
        result = pyhmmer.jackhmmer([], seqs, cpus=1, max_iterations=1)
        self.assertIs(None, next(result, None))

    def test_pksi(self):
        seqs = self.pksi
        results = pyhmmer.jackhmmer(seqs[-1:], seqs, cpus=1, max_iterations=1)
        result = next(results)
        self.assertEqual(result.iteration, 1)
        result.hits.sort()

        with self.table("A0A089QRB9.domtbl") as table:
            lines = iter(filter(lambda line: not line.startswith("#"), table))
            it = ((hit, domain) for hit in result.hits for domain in hit.domains)
            for line, (hit, domain) in itertools.zip_longest(lines, it):
                self.assertIsNot(line, None)
                self.assertIsNot(hit, None)
                fields = list(filter(None, line.strip().split(" ")))

                self.assertEqual(hit.name.decode(), fields[0])
                self.assertAlmostEqual(hit.score, float(fields[7]), delta=0.1)
                self.assertAlmostEqual(hit.bias, float(fields[8]), delta=0.1)
                self.assertAlmostEqual(hit.evalue, float(fields[6]), delta=0.1)

                self.assertAlmostEqual(domain.i_evalue, float(fields[12]), delta=0.1)
                self.assertAlmostEqual(domain.score, float(fields[13]), delta=0.1)

                self.assertEqual(domain.alignment.hmm_from, int(fields[15]))
                self.assertEqual(domain.alignment.hmm_to, int(fields[16]))
                self.assertEqual(domain.alignment.target_from, int(fields[17]))
                self.assertEqual(domain.alignment.target_to, int(fields[18]))
                self.assertEqual(domain.env_from, int(fields[19]))
                self.assertEqual(domain.env_to, int(fields[20]))

    def test_pksi_checkpoint(self):
        seqs = self.pksi
        # jackhmmer CLI converges in 3 iterations, 5 hits, 17 sequences in MSA
        iterations = next(
            pyhmmer.jackhmmer(
                seqs[-1],
                seqs,
                cpus=1,
                checkpoints=True,
            )
        )
        self.assertEqual(len(iterations), 3)
        self.assertTrue(iterations[-1].converged)
        self.assertEqual(len(iterations[-1].hits), 5)
        self.assertEqual(len(iterations[-1].msa.sequences), 17)

    def test_thioestherase(self):
        with self.hmm_file("Thioesterase") as hmm_file:
            hmm = hmm_file.read()
        with self.seqs_file("938293.PRJEB85.HG003687", digital=True) as seqs_file:
            seqs = seqs_file.read_block()

        result = next(
            pyhmmer.jackhmmer(
                hmm, seqs, cpus=1, max_iterations=1, incE=0.1, incdomE=0.1
            )
        )
        # unpack IterationResult
        _, hits, _, _, it = result
        self.assertEqual(it, 1)
        self.assertEqual(len(hits), 1)

        hits.sort()

        hit = hits[0]
        self.assertEqual(hit.name, b"938293.PRJEB85.HG003687_113")
        self.assertAlmostEqual(hit.score, 8.6, delta=0.1)  # printed with %6.1f
        self.assertAlmostEqual(hit.bias, 1.5, delta=0.1)  # printed with  %5.1f
        self.assertAlmostEqual(hit.evalue, 0.096, delta=0.01)  # printed with %9.2g
        self.assertEqual(len(hit.domains), 1)

        domain = hit.domains[0]
        self.assertAlmostEqual(domain.score, 8.1, delta=0.1)
        self.assertAlmostEqual(domain.bias, 1.5, delta=0.1)
        self.assertAlmostEqual(domain.i_evalue, 0.14, places=2)  # printed with %9.2g
        self.assertAlmostEqual(domain.c_evalue, 6.5e-05, places=2)  # printed with %9.2g
        self.assertEqual(domain.alignment.target_from, 115)
        self.assertEqual(domain.alignment.target_to, 129)
        self.assertEqual(domain.alignment.hmm_from, 79)
        self.assertEqual(domain.alignment.hmm_to, 93)
        self.assertEqual(domain.env_from, 115)
        self.assertEqual(domain.env_to, 129)

        last = hit.domains[-1]
        self.assertEqual(domain.alignment.hmm_name, last.alignment.hmm_name)


@unittest.skipUnless(resource_files, "importlib.resources not available")
class TestNhmmer(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        path = resource_files(__package__).joinpath("data", "hmms", "txt", "bmyD.hmm")
        with HMMFile(path) as hmm_file:
            cls.bmyD = next(hmm_file)
        path = resource_files(__package__).joinpath("data", "hmms", "txt", "RF00001.hmm")
        with HMMFile(path) as hmm_file:
            cls.rf00001 = next(hmm_file)

    @staticmethod
    def table(name):
        path = resource_files(__package__).joinpath("data", "tables", name)
        return path.open()

    def assertTableEqual(self, hits, table):
        lines = iter(filter(lambda line: not line.startswith("#"), table))
        reported_hits = filter(lambda hit: hit.reported, hits)
        for line, hit in itertools.zip_longest(lines, reported_hits):
            self.assertIsNot(line, None)
            self.assertIsNot(hit, None)
            fields = list(filter(None, line.strip().split(" ")))
            self.assertEqual(hit.name.decode(), fields[0])
            if fields[1] == "-":
                self.assertIs(hit.accession, None)
            else:
                self.assertEqual(hit.accession.decode(), fields[1])
            self.assertAlmostEqual(hit.best_domain.bias, float(fields[14]), delta=0.1)
            self.assertAlmostEqual(hit.best_domain.score, float(fields[13]), delta=0.1)
            self.assertAlmostEqual(
                hit.best_domain.i_evalue, float(fields[12]), delta=0.1
            )

    def test_no_queries(self):
        alphabet = Alphabet.dna()
        path = resource_files(__package__).joinpath("data", "seqs", "BGC0001090.gbk")
        with SequenceFile(path, digital=True, alphabet=alphabet) as seqs_file:
            seqs = seqs_file.read_block()
        hits = pyhmmer.nhmmer([], seqs, cpus=1)
        self.assertIs(None, next(hits, None))

    def test_bmyd_seq_bgc_block(self):
        alphabet = Alphabet.dna()

        path = resource_files(__package__).joinpath("data", "seqs", "bmyD.fna")
        with SequenceFile(path, digital=True, alphabet=alphabet) as seqs_file:
            query = next(seqs_file)

        path = resource_files(__package__).joinpath("data", "seqs", "BGC0001090.gbk")
        with SequenceFile(
            path, "genbank", digital=True, alphabet=alphabet
        ) as seqs_file:
            seqs = seqs_file.read_block()

        hits = next(pyhmmer.nhmmer(query, seqs, cpus=1))
        hits.sort()

        self.assertEqual(len(hits), 1)
        with self.table("bmyD3.tbl") as table:
            self.assertTableEqual(hits, table)

    def test_bmyd_seq_bgc_file(self):
        alphabet = Alphabet.dna()

        path = resource_files(__package__).joinpath("data", "seqs", "bmyD.fna")
        with SequenceFile(path, digital=True, alphabet=alphabet) as seqs_file:
            query = next(seqs_file)

        path = resource_files(__package__).joinpath("data", "seqs", "BGC0001090.gbk")
        with SequenceFile(path, "genbank", digital=True, alphabet=alphabet) as seqs:
            hits = list(pyhmmer.nhmmer(query, seqs, cpus=1))[0]
            hits.sort()

        self.assertEqual(len(hits.reported), 1)
        with self.table("bmyD3.tbl") as table:
            self.assertTableEqual(hits, table)

    def test_bmyd_msa_bgc_block(self):
        alphabet = Alphabet.dna()

        path = resource_files(__package__).joinpath("data", "seqs", "bmyD.fna")
        with SequenceFile(path, digital=True, alphabet=alphabet) as seqs_file:
            query = DigitalMSA(alphabet, name=b"bmyD", sequences=[next(seqs_file)])

        path = resource_files(__package__).joinpath("data", "seqs", "BGC0001090.gbk")
        with SequenceFile(
            path, "genbank", digital=True, alphabet=alphabet
        ) as seqs_file:
            seqs = seqs_file.read_block()

        hits = next(pyhmmer.nhmmer(query, seqs, cpus=1))
        self.assertEqual(len(hits), 1)

    def test_bmyd_msa_bgc_file(self):
        alphabet = Alphabet.dna()

        path = resource_files(__package__).joinpath("data", "seqs", "bmyD.fna")
        with SequenceFile(path, digital=True, alphabet=alphabet) as seqs_file:
            query = DigitalMSA(alphabet, name=b"bmyD", sequences=[next(seqs_file)])

        path = resource_files(__package__).joinpath("data", "seqs", "BGC0001090.gbk")
        with SequenceFile(path, "genbank", digital=True, alphabet=alphabet) as seqs:
            hits = list(pyhmmer.nhmmer(query, seqs, cpus=1))[0]
            self.assertEqual(len(hits.reported), 1)

    def test_bmyd_hmm_bgc_block(self):
        alphabet = Alphabet.dna()

        path = resource_files(__package__).joinpath("data", "seqs", "BGC0001090.gbk")
        with SequenceFile(
            path, "genbank", digital=True, alphabet=alphabet
        ) as seqs_file:
            seqs = seqs_file.read_block()

        hits = next(pyhmmer.nhmmer(self.bmyD, seqs, cpus=1))
        hits.sort()

        self.assertEqual(len(hits.reported), 2)
        with self.table("bmyD1.tbl") as table:
            self.assertTableEqual(hits, table)

    def test_bmyd_hmm_bgc_file(self):
        alphabet = Alphabet.dna()

        path = resource_files(__package__).joinpath("data", "seqs", "BGC0001090.gbk")
        with SequenceFile(
            path, "genbank", digital=True, alphabet=alphabet
        ) as seqs_file:
            hits = list(pyhmmer.nhmmer(self.bmyD, seqs_file, cpus=1))[0]
            hits.sort()

        self.assertEqual(len(hits.reported), 2)
        with self.table("bmyD1.tbl") as table:
            self.assertTableEqual(hits, table)

    def test_bmyd_hmm_genome_block(self):
        alphabet = Alphabet.dna()

        path = resource_files(__package__).joinpath("data", "seqs", "1390.SAMEA104415756.OFHT01000022.fna")
        with SequenceFile(path, "fasta", digital=True, alphabet=alphabet) as seqs_file:
            seqs = seqs_file.read_block()

        hits = next(pyhmmer.nhmmer(self.bmyD, seqs, cpus=1))
        hits.sort()

        self.assertEqual(len(hits.reported), 3)
        self.assertEqual(len(hits.included), 2)
        with self.table("bmyD2.tbl") as table:
            self.assertTableEqual(hits, table)

    def test_bmyd_hmm_genome_file(self):
        alphabet = Alphabet.dna()

        path = resource_files(__package__).joinpath("data", "seqs", "1390.SAMEA104415756.OFHT01000022.fna")
        with SequenceFile(path, "fasta", digital=True, alphabet=alphabet) as seqs_file:
            hits = list(pyhmmer.nhmmer(self.bmyD, seqs_file, cpus=1))[0]
            hits.sort()

        self.assertEqual(len(hits.reported), 3)
        self.assertEqual(len(hits.included), 2)
        with self.table("bmyD2.tbl") as table:
            self.assertTableEqual(hits, table)

    def test_rf0001_genome_file(self):
        alphabet = Alphabet.rna()
        path = resource_files(__package__).joinpath("data", "seqs", "1390.SAMEA104415756.OFHT01000024.fna")
        with SequenceFile(path, "fasta", digital=True, alphabet=alphabet) as seqs_file:
            hits = list(pyhmmer.nhmmer(self.rf00001, seqs_file, cpus=1))[0]
            hits.sort()

        self.assertEqual(len(hits), 1)
        self.assertAlmostEqual(hits[0].evalue, 2.5e-17, places=2)
        self.assertEqual(hits[0].best_domain.strand, "-")

    def test_rf0001_genome_file_wlen_3878(self):
        alphabet = Alphabet.rna()
        path = resource_files(__package__).joinpath("data", "seqs", "1390.SAMEA104415756.OFHT01000024.fna")
        with SequenceFile(path, "fasta", digital=True, alphabet=alphabet) as seqs_file:
            hits = list(pyhmmer.nhmmer(self.rf00001, seqs_file, cpus=1, window_length=3878))[0]
            hits.sort()

        self.assertEqual(len(hits), 2)
        self.assertAlmostEqual(hits[0].evalue, 5.4e-17, places=2)
        self.assertAlmostEqual(hits[1].evalue, 0.3, places=2)
        self.assertEqual(hits[0].best_domain.strand, "-")
        self.assertEqual(hits[1].best_domain.strand, "-")


class TestHmmalign(unittest.TestCase):
    def setUp(self):
        self.tmpout = tempfile.NamedTemporaryFile(suffix=".hmm", delete=False).name

    def tearDown(self):
        os.remove(self.tmpout)

    @unittest.skipUnless(resource_files, "importlib.resources not available")
    def test_luxc(self):
        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "LuxC.hmm")
        seqs_path = resource_files(__package__).joinpath("data", "seqs", "LuxC.faa")
        ref_path = resource_files(__package__).joinpath("data", "msa", "LuxC.hmmalign.sto")

        with HMMFile(hmm_path) as hmm_file:
            hmm = hmm_file.read()
        with SequenceFile(seqs_path, digital=True, alphabet=hmm.alphabet) as seqs_file:
            seqs = seqs_file.read_block()
        with MSAFile(ref_path) as ref_file:
            ref = ref_file.read()

        msa = pyhmmer.hmmalign(hmm, seqs, trim=True, cpus=1)
        self.assertEqual(msa, ref)


class TestHMMScan(unittest.TestCase):
    @staticmethod
    def table(name):
        path = resource_files(__package__).joinpath("data", "tables", name)
        return path.open()

    @staticmethod
    def hmm_file(name):
        path = resource_files(__package__).joinpath("data", "hmms", "db", "{}.hmm".format(name))
        return HMMFile(path)

    @unittest.skipUnless(resource_files, "importlib.resources not available")
    def test_rrefam_block(self):
        # get paths to resources
        table_path = resource_files(__package__).joinpath("data", "tables", "RREFam.scan.tbl")
        db_path = resource_files(__package__).joinpath("data", "hmms", "db", "RREFam.hmm")
        seqs_path = resource_files(__package__).joinpath("data", "seqs", "PKSI.faa")

        # load expected results from the hmmscan table
        expected = {}
        with open(table_path) as table:
            lines = filter(lambda line: not line.startswith("#"), table)
            for query_name, query_lines in itertools.groupby(
                lines, key=lambda line: line.strip().split()[2]
            ):
                expected[query_name] = list(query_lines)

        # pre-load the profile database so that `hmmscan` will create a profile block
        with HMMFile(db_path) as hmm_file:
            hmms = list(hmm_file)

        # scan with the sequences and check the hits are equal
        with SequenceFile(
            seqs_path, digital=True, alphabet=hmms[0].alphabet
        ) as seqs_file:
            for hits in pyhmmer.hmmer.hmmscan(seqs_file, hmms, cpus=1):
                expected_lines = expected.get(hits.query.name.decode())
                if expected_lines is None:
                    self.assertEqual(len(hits), 0)
                    continue
                for line, hit in itertools.zip_longest(expected_lines, hits):
                    self.assertIsNot(line, None)
                    self.assertIsNot(hit, None)
                    fields = list(filter(None, line.strip().split(" ")))
                    self.assertEqual(hit.name.decode(), fields[0])
                    self.assertAlmostEqual(hit.score, float(fields[5]), delta=0.1)
                    self.assertAlmostEqual(hit.bias, float(fields[6]), delta=0.1)
                    self.assertAlmostEqual(hit.evalue, float(fields[4]), delta=0.1)

    @unittest.skipUnless(resource_files, "importlib.resources not available")
    def test_rrefam_file(self):
        # get paths to resources
        table_path = resource_files(__package__).joinpath("data", "tables", "RREFam.scan.tbl")
        db_path = resource_files(__package__).joinpath("data", "hmms", "db", "RREFam.hmm")
        seqs_path = resource_files(__package__).joinpath("data", "seqs", "PKSI.faa")

        # load expected results from the hmmscan table
        expected = {}
        with open(table_path) as table:
            lines = filter(lambda line: not line.startswith("#"), table)
            for query_name, query_lines in itertools.groupby(
                lines, key=lambda line: line.strip().split()[2]
            ):
                expected[query_name] = list(query_lines)

        # scan with the sequences and check the hits are equal, using the file
        # to read the profiles from (the files will be rewinded before each query)
        with SequenceFile(seqs_path, digital=True) as seqs_file:
            with HMMPressedFile(db_path) as pressed_file:
                for hits in pyhmmer.hmmer.hmmscan(seqs_file, pressed_file, cpus=1):
                    expected_lines = expected.get(hits.query.name.decode())
                    if expected_lines is None:
                        self.assertEqual(len(hits), 0)
                        continue
                    for line, hit in itertools.zip_longest(expected_lines, hits):
                        self.assertIsNot(line, None)
                        self.assertIsNot(hit, None)
                        fields = list(filter(None, line.strip().split(" ")))
                        self.assertEqual(hit.name.decode(), fields[0])
                        self.assertAlmostEqual(hit.score, float(fields[5]), delta=0.1)
                        self.assertAlmostEqual(hit.bias, float(fields[6]), delta=0.1)
                        self.assertAlmostEqual(hit.evalue, float(fields[4]), delta=0.1)



from libc.stdint cimport uint8_t


cdef extern from "p7_config.h" nogil:

    const double p7_ETARGET_AMINO
    const double p7_ETARGET_DNA
    const double p7_ETARGET_OTHER

    const size_t p7_MAXABET
    const size_t p7_MAXCODE
    const size_t p7_MAX_SC_TXTLEN
    const size_t p7_MAXDCHLET


cdef extern from "hmmer.h" nogil:

    # Search modes
    cdef enum:
        p7_NO_MODE   = 0
        p7_LOCAL     = 1
        p7_GLOCAL    = 2
        p7_UNILOCAL  = 3
        p7_UNIGLOCAL = 4

    cdef bint p7_IsLocal(int)
    cdef bint p7_IsMulti(int)


    const size_t p7_NEVPARAM
    cdef enum p7_evparams_e:
        p7_MMU     = 0
        p7_MLAMBDA = 1
        p7_VMU     = 2
        p7_VLAMBDA = 3
        p7_FTAU    = 4
        p7_FLAMBDA = 5

    const float p7_EVPARAM_UNSET


    const size_t p7_NCUTOFFS
    cdef enum p7_cutoffs_e:
        p7_GA1 = 0
        p7_GA2 = 1
        p7_TC1 = 2
        p7_TC2 = 3
        p7_NC1 = 4
        p7_NC2 = 5

    const float p7_CUTOFF_UNSET


    const size_t p7_NOFFSETS
    cdef enum p7_offsets_e:
        p7_MOFFSET = 0
        p7_FOFFSET = 1
        p7_POFFSET = 2

    const int p7_DEFAULT
    const int p7_DIGITIZE
    const int p7_ALL_CONSENSUS_COLS
    const int p7_TRIM



# coding: utf-8
# cython: language_level=3
"""Common errors and status codes for the `easel` and `hmmer` modules.
"""

cimport libeasel
cimport libeasel.alphabet

include "_getid.pxi"

statuscode = {
    libeasel.eslOK: "eslOK",
    libeasel.eslFAIL: "eslFAIL",
    libeasel.eslEOL: "eslEOL",
    libeasel.eslEOF: "eslEOF",
    libeasel.eslEOD: "eslEOD",
    libeasel.eslEMEM: "eslEMEM",
    libeasel.eslENOTFOUND: "eslENOTFOUND",
    libeasel.eslEFORMAT: "eslEFORMAT",
    libeasel.eslEAMBIGUOUS: "eslEAMBIGUOUS",
    libeasel.eslEINCOMPAT: "eslEINCOMPAT",
    libeasel.eslEINVAL: "eslEINVAL",
    libeasel.eslESYS: "eslESYS",
    libeasel.eslECORRUPT: "eslECORRUPT",
    libeasel.eslEINCONCEIVABLE: "eslEINCONCEIVABLE",
    libeasel.eslESYNTAX: "eslESYNTAX",
    libeasel.eslERANGE: "eslERANGE",
    libeasel.eslEDUP: "eslEDUP",
    libeasel.eslENORESULT: "eslENORESULT",
    libeasel.eslETYPE: "eslETYPE",
    libeasel.eslEOVERWRITE: "eslEOVERWRITE",
    libeasel.eslENOSPACE: "eslENOSPACE",
    libeasel.eslEUNIMPLEMENTED: "eslEUNIMPLEMENTED",
    libeasel.eslENOFORMAT: "eslENOFORMAT",
    libeasel.eslENOALPHABET: "eslENOALPHABET",
    libeasel.eslEWRITE: "eslEWRITE",
    libeasel.eslEINACCURATE: "eslEINACCURATE",
}


class UnexpectedError(RuntimeError):
    """An unexpected error that happened in the C code.

    As a user of this library, you should never see this exception being
    raised. If you do, please open an issue with steps to reproduce on the
    `bug tracker <https://github.com/althonos/pyhmmer/issues>`_, so that
    proper error handling can be added to the relevant part of the bindings.

    """

    def __init__(self, int code, str function):
        """__init__(self, code, function)\n--\n
        """
        self.code = code
        self.function = function

    def __repr__(self):
        return "{}({!r}, {!r})".format(type(self).__name__, self.code, self.function)

    def __str__(self):
        return "Unexpected error occurred in {!r}: {} (status code {})".format(
            self.function,
            statuscode.get(self.code, "<unknown>"),
            self.code,
        )


class AllocationError(MemoryError):
    """A memory error that is caused by an unsuccessful allocation.
    """

    def __init__(self, str ctype, size_t itemsize, size_t count=1):
        """__init__(self, ctype, itemsize, count=1)\n--\n
        """
        self.ctype = ctype
        self.itemsize = itemsize
        self.count = count

    def __repr__(self):
        cdef str ty = type(self).__name__
        if self.count == 1:
            return f"{ty}({self.ctype!r}, {self.itemsize})"
        return f"{ty}({self.ctype!r}, {self.itemsize}, {self.count})"

    def __str__(self):
        if self.count == 1:
            return f"Could not allocate {self.itemsize} bytes for type {self.ctype}"
        return f"Could not allocate {self.itemsize*self.count} bytes for an array of {self.count} {self.ctype}"


class EaselError(RuntimeError):
    """An error that was raised from the Easel code.
    """

    def __init__(self, int code, str message):
        """__init__(self, code, message)\n--\n
        """
        self.code = code
        self.message = message

    def __repr__(self):
        cdef str ty = type(self).__name__
        return f"{ty}({self.code!r}, {self.message!r})"

    def __str__(self):
        return "Error raised from C code: {}, {} (status code {})".format(
            self.message,
            statuscode.get(self.code, "<unknown>"),
            self.code
        )


class AlphabetMismatch(ValueError):
    """A value error caused by an alphabet mismatch.

    This error is raised on occasions where several arguments to a
    function have mismatching alphabets. For instance, passing a
    `~pyhmmer.easel.DigitalSequence` to a `~pyhmmer.plan7.Builder` with
    different alphabets::

        >>> seq = easel.DigitalSequence(easel.Alphabet.dna())
        >>> builder = plan7.Builder(easel.Alphabet.amino())
        >>> builder.build(seq, plan7.Background(seq.alphabet))
        Traceback (most recent call last):
          ...
        pyhmmer.errors.AlphabetMismatch: Expected amino alphabet, found DNA alphabet

    .. versionadded:: 0.3.0

    """

    def __init__(self, expected, actual=None):
        """__init__(self, expected, actual=None)\n--\n
        """
        super().__init__(self)
        self.expected = expected
        self.actual = actual

    def __repr__(self):
        cdef str ty = type(self).__name__
        if self.actual is None:
            return f"{ty}({self.expected!r})"
        return f"{ty}({self.expected!r}, {self.actual!r})"

    def __str__(self):
        if self.actual is None:
            return f"Expected {self.expected.type} alphabet"
        return f"Expected {self.expected.type} alphabet, found {self.actual.type} alphabet"

    def __eq__(self, other):
        if not isinstance(other, AlphabetMismatch):
            return NotImplemented
        return self.actual == other.actual and self.expected == other.expected


class ServerError(RuntimeError):
    """A runtime error that happened in a ``hmmpgmd`` server.
    """

    def __init__(self, int code, str message):
        """__init__(self, code, message)\n--\n
        """
        self.code = code
        self.message = message

    def __repr__(self):
        cdef str ty = type(self).__name__
        return f"{ty}({self.code!r}, {self.message!r})"

    def __str__(self):
        return "Error in server: {}, {} (status code {})".format(
            self.message,
            statuscode.get(self.code, "<unknown>"),
            self.code,
        )


class MissingCutoffs(ValueError):
    """The model is missing bitscore cutoffs required by the pipeline.

    .. versionadded:: 0.8.2

    .. versionadded:: 0.10.8
       The ``model_name`` and ``bit_cutoffs`` displayed in the error message.

    """

    def __init__(self, str model_name = None, str bit_cutoffs = None):
        """__init__(self, model_name=None, bit_cutoffs=None)\n--\n
        """
        self.model_name = model_name
        self.bit_cutoffs = bit_cutoffs

    def __repr__(self):
        cdef str ty = type(self).__name__
        args = []
        if self.model_name is not None:
            args.append(f"model_name={self.model_name!r}")
        if self.bit_cutoffs is not None:
            args.append(f"bit_cutoffs={self.bit_cutoffs!r}")
            return f"{ty}({self.model_name!r})"
        return f"{ty}({ ', '.join(args) })"

    def __str__(self):
        model = f"Model {self.model_name!r}" if self.model_name is not None else "Model"
        bit_cutoffs = f"{self.bit_cutoffs!r} bitscore cutoff" if self.bit_cutoffs is not None else f"bitscore cutoff"
        return f"{model} is missing {bit_cutoffs} required by pipeline"


class InvalidParameter(ValueError):
    """An invalid parameter value was given.

    .. versionadded:: 0.8.2

    """

    def __init__(self, str name, object value, *, list choices=None, str hint=None):
        """__init__(self, name, value, *, choices=None, hint=None)\n--\n
        """
        self.name = name
        self.value = value
        self.choices = choices
        self.hint = hint

    def __repr__(self):
        cdef str ty = type(self).__name__
        return f"{ty}({self.name!r}, {self.value!r}, {self.choices!r})"

    def __str__(self):
        options = ""
        hint = self.hint

        if hint is None and self.choices is not None:
            choices = [
                x.__name__
                if isinstance(x, type)
                else repr(x)
                for x in self.choices
            ]
            hint = " or ".join([
                ", ".join(choices[:len(self.choices)-1]),
                choices[len(self.choices)-1]
            ])
        if hint is not None:
            options = f" (expected {hint})"

        return f"Invalid {self.name!r} parameter value: {self.value!r}{options}"


class InvalidHMM(ValueError):
    """A value error caused by a HMM that fails validation.

    .. versionadded:: 0.10.3

    """

    def __init__(self, object hmm, str message):
        """__init__(self, hmm, message)\n--\n
        """
        super().__init__(hmm, message)
        self.message = message
        self.hmm = hmm

    def __repr__(self):
        cdef str ty = type(self).__name__
        return f"{ty}({self.code!r}, {self.message!r})"

    def __str__(self):
        return "HMM {!r} is invalid: {}".format(
            self.hmm.name.decode('utf-8', 'replace'),
            self.message
        )




import array
import unittest
import pickle
import struct
import sys

from pyhmmer.easel import Vector, VectorD, VectorF, VectorU8


class _TestVectorBase(object):

    Vector = NotImplemented

    def test_pickle(self):
        v1 = self.Vector(range(6))
        v2 = pickle.loads(pickle.dumps(v1))
        self.assertSequenceEqual(v1, v2)

    def test_pickle_protocol4(self):
        v1 = self.Vector(range(6))
        v2 = pickle.loads(pickle.dumps(v1, protocol=4))
        self.assertEqual(v1.shape, v2.shape)
        self.assertSequenceEqual(v1, v2)
        self.assertSequenceEqual(memoryview(v1), memoryview(v2))

    @unittest.skipUnless(sys.version_info >= (3, 8), "pickle protocol 5 requires Python 3.8+")
    def test_pickle_protocol5(self):
        v1 = self.Vector(range(6))
        v2 = pickle.loads(pickle.dumps(v1, protocol=5))
        self.assertEqual(v1.shape, v2.shape)
        self.assertSequenceEqual(v1, v2)
        self.assertSequenceEqual(memoryview(v1), memoryview(v2))

    def test_empty_vector(self):
        v1 = self.Vector([])
        v2 = self.Vector.zeros(0)
        v3 = self.Vector()
        self.assertEqual(len(v1), 0)
        self.assertEqual(len(v2), 0)
        self.assertEqual(len(v3), 0)
        self.assertEqual(v1, v2)
        self.assertEqual(v1, v3)
        self.assertFalse(v1)
        self.assertFalse(v2)
        self.assertFalse(v3)

        if sys.implementation.name != "pypy":
            v3 = self.Vector.zeros(3)
            self.assertLess(sys.getsizeof(v1), sys.getsizeof(v3))

    def test_init(self):
        vec = self.Vector([1, 2, 3])
        self.assertEqual(vec[0], 1)
        self.assertEqual(vec[1], 2)
        self.assertEqual(vec[2], 3)

    def test_init_memcpy(self):
        v1 = self.Vector([1, 2, 3])
        a  = array.array(v1.format, v1)
        v2 = self.Vector(a)
        self.assertEqual(v1, v2)

    def test_init_error(self):
        self.assertRaises(TypeError, self.Vector, 1)
        self.assertRaises(TypeError, self.Vector.zeros, [1, 2, 3])
        self.assertRaises(TypeError, self.Vector.zeros, "1")

    def test_shape(self):
        vec = self.Vector([1, 2, 3])
        self.assertEqual(vec.shape, (3,))
        vec2 = self.Vector.zeros(100)
        self.assertEqual(vec2.shape, (100,))
        vec3 = self.Vector.zeros(0)
        self.assertEqual(vec3.shape, (0,))

    def test_len(self):
        vec = self.Vector([1, 2, 3])
        self.assertEqual(len(vec), 3)
        vec2 = self.Vector.zeros(100)
        self.assertEqual(len(vec2), 100)
        vec3 = self.Vector([])
        self.assertEqual(len(vec3), 0)

    def test_copy(self):
        vec = self.Vector([1, 2, 3])
        vec2 = vec.copy()
        del vec
        self.assertIsInstance(vec2, self.Vector)
        self.assertEqual(vec2[0], 1)
        self.assertEqual(vec2[1], 2)
        self.assertEqual(vec2[2], 3)

        vec3 = self.Vector([])
        vec4 = vec3.copy()
        self.assertEqual(vec3, vec4)
        self.assertEqual(len(vec4), 0)

    def test_reverse(self):
        vec = self.Vector([1, 2, 3])
        vec.reverse()
        self.assertEqual(vec[0], 3)
        self.assertEqual(vec[1], 2)
        self.assertEqual(vec[2], 1)

        vec2 = self.Vector([1, 2, 3, 4])
        vec2.reverse()
        self.assertEqual(vec2[0], 4)
        self.assertEqual(vec2[1], 3)
        self.assertEqual(vec2[2], 2)
        self.assertEqual(vec2[3], 1)

        vec3 = self.Vector([])
        vec3.reverse()
        self.assertEqual(vec3, self.Vector([]))
        self.assertEqual(len(vec3), 0)

    def test_add(self):
        vec = self.Vector([1, 2, 3])
        vec2 = vec + 1
        self.assertEqual(vec2[0], 2)
        self.assertEqual(vec2[1], 3)
        self.assertEqual(vec2[2], 4)

        with self.assertRaises(ValueError):
            vec + self.Vector([1])

        v2 = self.Vector([])
        v3 = v2 + self.Vector([])
        self.assertEqual(v3, self.Vector([]))

    def test_iadd_scalar(self):
        vec = self.Vector([1, 2, 3])
        vec += 3
        self.assertEqual(vec[0], 4)
        self.assertEqual(vec[1], 5)
        self.assertEqual(vec[2], 6)

        v2 = self.Vector([])
        v2 += 3
        self.assertEqual(v2, self.Vector([]))

    def test_iadd_vector(self):
        vec = self.Vector([4, 5, 6])
        vec += self.Vector([10, 11, 12])
        self.assertEqual(vec[0], 14)
        self.assertEqual(vec[1], 16)
        self.assertEqual(vec[2], 18)

        with self.assertRaises(ValueError):
            vec += self.Vector([1])

        v2 = self.Vector([])
        v2 += self.Vector([])
        self.assertEqual(v2, self.Vector([]))

    def test_sub(self):
        vec = self.Vector([1, 2, 3])
        v2 = vec - 1
        self.assertEqual(v2[0], 0)
        self.assertEqual(v2[1], 1)
        self.assertEqual(v2[2], 2)

        v3 = self.Vector([8, 10, 12])
        v4 = self.Vector([1, 2, 3])
        v5 = v3 - v4
        self.assertEqual(v5[0], 7)
        self.assertEqual(v5[1], 8)
        self.assertEqual(v5[2], 9)

    def test_isub_scalar(self):
        vec = self.Vector([4, 5, 6])
        vec -= 2
        self.assertEqual(vec[0], 2)
        self.assertEqual(vec[1], 3)
        self.assertEqual(vec[2], 4)

    def test_isub_vector(self):
        vec = self.Vector([4, 5, 6])
        vec -= self.Vector([2, 3, 2])
        self.assertEqual(vec[0], 2)
        self.assertEqual(vec[1], 2)
        self.assertEqual(vec[2], 4)

    def test_mul_scalar(self):
        vec = self.Vector([1, 2, 3])
        v2 = vec * 3
        self.assertEqual(v2[0], 3)
        self.assertEqual(v2[1], 6)
        self.assertEqual(v2[2], 9)

        v2 = self.Vector([])
        v3 = v2 * 3
        self.assertEqual(v3, self.Vector([]))

    def test_mul_vector(self):
        vec = self.Vector([1, 2, 3])
        v2 = self.Vector([3, 6, 9])
        v3 = vec * v2
        self.assertEqual(v3[0], 3)
        self.assertEqual(v3[1], 12)
        self.assertEqual(v3[2], 27)

        v2 = self.Vector([])
        v3 = v2 * self.Vector([])
        self.assertEqual(v3, self.Vector([]))

    def test_imul_scalar(self):
        vec = self.Vector([1, 2, 3])
        vec *= 3
        self.assertEqual(vec[0], 3)
        self.assertEqual(vec[1], 6)
        self.assertEqual(vec[2], 9)

        v2 = self.Vector([])
        v2 *= 3
        self.assertEqual(v2, self.Vector([]))

    def test_matmul_vector(self):
        u = self.Vector([4, 5, 6])
        v = self.Vector([1, 2, 3])
        self.assertEqual(u @ v, 1*4 + 2*5 + 3*6)

        x = self.Vector([])
        y = self.Vector([])
        self.assertEqual(x @ y, 0)

    def test_sum(self):
        vec = self.Vector([1, 2, 3])
        self.assertEqual(vec.sum(), 1 + 2 + 3)

        vec2 = self.Vector([])
        self.assertEqual(vec2.sum(), 0)

    def test_slice(self):
        vec = self.Vector([1, 2, 3, 4])

        v1 = vec[:]
        self.assertEqual(len(v1), 4)
        self.assertEqual(v1[0], 1)
        self.assertEqual(v1[-1], 4)

        v2 = vec[1:3]
        self.assertEqual(len(v2), 2)
        self.assertEqual(v2[0], 2)
        self.assertEqual(v2[1], 3)

        v3 = vec[:-1]
        self.assertEqual(len(v3), 3)
        self.assertEqual(v3[-1], 3)

        v4 = vec[0:10]
        self.assertEqual(len(v4), 4)
        self.assertEqual(v4[-1], 4)

        with self.assertRaises(ValueError):
            vec[::-1]

    def test_min(self):
        vec = self.Vector([1, 2, 3])
        self.assertEqual(vec.min(), 1)

        v2 = self.Vector([])
        self.assertRaises(ValueError, v2.min)

    def test_max(self):
        vec = self.Vector([1, 2, 3])
        self.assertEqual(vec.max(), 3)

        v2 = self.Vector([])
        self.assertRaises(ValueError, v2.max)

    def test_argmin(self):
        vec = self.Vector([4, 2, 8])
        self.assertEqual(vec.argmin(), 1)

        v2 = self.Vector([])
        self.assertRaises(ValueError, v2.argmin)

    def test_argmax(self):
        vec = self.Vector([2, 8, 4])
        self.assertEqual(vec.argmax(), 1)

        v2 = self.Vector([])
        self.assertRaises(ValueError, v2.argmax)


class TestVector(unittest.TestCase):

    def test_abstract(self):
        self.assertRaises(TypeError, Vector, [1, 2, 3])
        self.assertRaises(TypeError, Vector.zeros, 1)


class TestVectorF(_TestVectorBase, unittest.TestCase):

    Vector = VectorF

    def test_strides(self):
        vec = self.Vector([1, 2, 3])
        sizeof_float = len(struct.pack('f', 1.0))
        self.assertEqual(vec.strides, (sizeof_float,))

    def test_normalize(self):
        vec = self.Vector([1, 3])
        vec.normalize()
        self.assertEqual(vec[0], 1/4)
        self.assertEqual(vec[1], 3/4)

        vec2 = self.Vector([])
        vec2.normalize()

    def test_memoryview_tolist(self):
        vec = self.Vector([1, 2, 3])
        mem = memoryview(vec)
        self.assertEqual(mem.tolist(), [1.0, 2.0, 3.0])

    def test_neg(self):
        vec = self.Vector([1, 2, 3])
        v2 = -vec
        self.assertEqual(v2[0], -1)
        self.assertEqual(v2[1], -2)
        self.assertEqual(v2[2], -3)

    def test_div_scalar(self):
        vec = self.Vector([1, 2, 3])
        v2 = vec / 2
        self.assertEqual(v2[0], 0.5)
        self.assertEqual(v2[1], 1.0)
        self.assertEqual(v2[2], 1.5)

        v2 = self.Vector([])
        v3 = v2 / 3
        self.assertEqual(v3, self.Vector([]))

    def test_div_vector(self):
        vec = self.Vector([1, 2, 3])
        v2 = self.Vector([2, 4, 6])
        v3 = vec / v2
        self.assertEqual(v3[0], 0.5)
        self.assertEqual(v3[1], 0.5)
        self.assertEqual(v3[2], 0.5)

        v2 = self.Vector([])
        v3 = v2 / self.Vector([])
        self.assertEqual(v3, self.Vector([]))

    def test_idiv_scalar(self):
        vec = self.Vector([1, 2, 3])
        vec /= 2
        self.assertEqual(vec[0], 0.5)
        self.assertEqual(vec[1], 1.0)
        self.assertEqual(vec[2], 1.5)

        vec = self.Vector([])
        vec /= 3
        self.assertEqual(vec, self.Vector([]))

    def test_idiv_vector(self):
        vec = self.Vector([1, 2, 3])
        vec /= self.Vector([2, 4, 6])
        self.assertEqual(vec[0], 0.5)
        self.assertEqual(vec[1], 0.5)
        self.assertEqual(vec[2], 0.5)

        vec = self.Vector([])
        vec /= self.Vector([])
        self.assertEqual(vec, self.Vector([]))

    def test_from_raw_bytes_littleendian(self):
        vec = self.Vector._from_raw_bytes(b'\x00\x00\x80>', 1, byteorder="little")
        self.assertEqual(len(vec), 1)
        self.assertEqual(vec[0], 0.25)

    def test_from_raw_bytes_bigendian(self):
        vec = self.Vector._from_raw_bytes(b'>\x80\x00\x00', 1, byteorder="big")
        self.assertEqual(len(vec), 1)
        self.assertEqual(vec[0], 0.25)


class TestVectorD(_TestVectorBase, unittest.TestCase):

    Vector = VectorD

    def test_strides(self):
        vec = self.Vector([1, 2, 3])
        sizeof_double = len(struct.pack('d', 1.0))
        self.assertEqual(vec.strides, (sizeof_double,))

    def test_normalize(self):
        vec = self.Vector([1, 3])
        vec.normalize()
        self.assertEqual(vec[0], 1/4)
        self.assertEqual(vec[1], 3/4)

        vec2 = self.Vector([])
        vec2.normalize()

    def test_memoryview_tolist(self):
        vec = self.Vector([1, 2, 3])
        mem = memoryview(vec)
        self.assertEqual(mem.tolist(), [1.0, 2.0, 3.0])

    def test_neg(self):
        vec = self.Vector([1, 2, 3])
        v2 = -vec
        self.assertEqual(v2[0], -1)
        self.assertEqual(v2[1], -2)
        self.assertEqual(v2[2], -3)

    def test_div_scalar(self):
        vec = self.Vector([1, 2, 3])
        v2 = vec / 2
        self.assertEqual(v2[0], 0.5)
        self.assertEqual(v2[1], 1.0)
        self.assertEqual(v2[2], 1.5)

        v2 = self.Vector([])
        v3 = v2 / 3
        self.assertEqual(v3, self.Vector([]))

    def test_div_vector(self):
        vec = self.Vector([1, 2, 3])
        v2 = self.Vector([2, 4, 6])
        v3 = vec / v2
        self.assertEqual(v3[0], 0.5)
        self.assertEqual(v3[1], 0.5)
        self.assertEqual(v3[2], 0.5)

        v2 = self.Vector([])
        v3 = v2 / self.Vector([])
        self.assertEqual(v3, self.Vector([]))

    def test_idiv_scalar(self):
        vec = self.Vector([1, 2, 3])
        vec /= 2
        self.assertEqual(vec[0], 0.5)
        self.assertEqual(vec[1], 1.0)
        self.assertEqual(vec[2], 1.5)

        vec = self.Vector([])
        vec /= 3
        self.assertEqual(vec, self.Vector([]))

    def test_idiv_vector(self):
        vec = self.Vector([1, 2, 3])
        vec /= self.Vector([2, 4, 6])
        self.assertEqual(vec[0], 0.5)
        self.assertEqual(vec[1], 0.5)
        self.assertEqual(vec[2], 0.5)

        vec = self.Vector([])
        vec /= self.Vector([])
        self.assertEqual(vec, self.Vector([]))

    def test_from_raw_bytes_littleendian(self):
        vec = self.Vector._from_raw_bytes(b'\x00\x00\x00\x00\x00\x00\xd0\x3f', 1, byteorder="little")
        self.assertEqual(len(vec), 1)
        self.assertEqual(vec[0], 0.25)

    def test_from_raw_bytes_bigendian(self):
        vec = self.Vector._from_raw_bytes(b'\x3f\xd0\x00\x00\x00\x00\x00\x00', 1, byteorder="big")
        self.assertEqual(len(vec), 1)
        self.assertEqual(vec[0], 0.25)


class TestVectorU8(_TestVectorBase, unittest.TestCase):

    Vector = VectorU8

    def test_strides(self):
        vec = self.Vector([1, 2, 3])
        sizeof_u8 = len(struct.pack('B', 1))
        self.assertEqual(vec.strides, (sizeof_u8,))

    def test_isub_wrapping(self):
        vec = self.Vector([0, 1, 2])
        vec -= 1
        self.assertEqual(vec[0], 255)
        self.assertEqual(vec[1], 0)
        self.assertEqual(vec[2], 1)

    def test_sum_wrapping(self):
        vec = self.Vector([124, 72, 116])
        self.assertEqual(vec.sum(), (124 + 72 + 116) % 256)

    def test_memoryview_tolist(self):
        vec = self.Vector([1, 2, 3])
        mem = memoryview(vec)
        self.assertEqual(mem.tolist(), [1, 2, 3])

    def test_eq_bytebuffer(self):
        vec = self.Vector([1, 2, 3])
        b1 = bytearray([1, 2, 3])
        self.assertEqual(vec, b1)

        b2 = array.array('B', [1, 2, 3])
        self.assertEqual(vec, b2)

        b3 = array.array('B', [1, 2, 3, 4])
        self.assertNotEqual(vec, b3)

        b4 = array.array('L', [1, 2, 3])
        self.assertNotEqual(vec, b4)

    def test_floordiv_scalar(self):
        vec = self.Vector([1, 2, 3])
        v2 = vec // 2
        self.assertEqual(v2[0], 0)
        self.assertEqual(v2[1], 1)
        self.assertEqual(v2[2], 1)

        v2 = self.Vector([])
        v3 = v2 // 3
        self.assertEqual(v3, self.Vector([]))

    def test_floordiv_vector(self):
        vec = self.Vector([1, 2, 3])
        v2 = self.Vector([2, 4, 1])
        v3 = vec // v2
        self.assertEqual(v3[0], 0)
        self.assertEqual(v3[1], 0)
        self.assertEqual(v3[2], 3)

        v2 = self.Vector([])
        v3 = v2 // self.Vector([])
        self.assertEqual(v3, self.Vector([]))

    def test_ifloordiv_scalar(self):
        vec = self.Vector([1, 2, 3])
        vec //= 2
        self.assertEqual(vec[0], 0)
        self.assertEqual(vec[1], 1)
        self.assertEqual(vec[2], 1)

        vec = self.Vector([])
        vec //= 3
        self.assertEqual(vec, self.Vector([]))

    def test_ifloordiv_vector(self):
        vec = self.Vector([1, 2, 3])
        vec //= self.Vector([2, 4, 6])
        self.assertEqual(vec[0], 0)
        self.assertEqual(vec[1], 0)
        self.assertEqual(vec[2], 0)

        vec = self.Vector([])
        vec //= self.Vector([])
        self.assertEqual(vec, self.Vector([]))



import copy
import pickle
import unittest
import sys

from pyhmmer.easel import Randomness


class TestRandomness(unittest.TestCase):

    @unittest.skipIf(sys.implementation.name == "pypy", "`getsizeof` not supported on PyPY")
    def test_sizeof(self):
        rng = Randomness(42, fast=True)
        self.assertGreater(sys.getsizeof(rng), 0)

    def test_init_fast(self):
        rng = Randomness(42, fast=True)
        self.assertTrue(rng.fast)

    def test_init_mersenne(self):
        rng = Randomness(42)
        self.assertFalse(rng.fast)

    def test_init_error(self):
        self.assertRaises(TypeError, Randomness, "ok")
        self.assertRaises(OverflowError, Randomness, -1)

    def test_copy(self):
        rng = Randomness(42)
        rng.random() # advance the rng
        new = copy.copy(rng)
        self.assertEqual(new.random(), rng.random())

    def test_pickle(self):
        rng = Randomness(42)
        rng.random() # advance the rng
        new = pickle.loads(pickle.dumps(rng))
        self.assertEqual(new.random(), rng.random())

    def test_normalvariate(self):
        rng1 = Randomness(42)
        n1 = rng1.normalvariate(0, 1)
        rng2 = Randomness(42)
        n2 = rng2.normalvariate(0, 1)
        self.assertEqual(n1, n2)

    def test_random(self):
        rng1 = Randomness(42)
        n1 = rng1.random()
        rng2 = Randomness(42)
        n2 = rng2.random()
        self.assertEqual(n1, n2)



#! /usr/bin/env python3

import sys
import os
import subprocess

usage = 'pmark-master.py <top_builddir> <top_srcdir> <resultdir> <ncpu> <benchmark_pfx> <pmark_script>'

if len(sys.argv) != 7: sys.exit('Incorrect number of cmdline args.\nUsage: {}'.format(usage))

(top_builddir, top_srcdir, resultdir, ncpu, benchmark_pfx, pmark_script) = sys.argv[1:]

tblfile = benchmark_pfx + '.tbl'
msafile = benchmark_pfx + '.train.msa'
fafile  = benchmark_pfx + '.test.fa'

if     os.path.exists(resultdir):        sys.exit('results directory {} already exists'.format(resultdir))
if not os.path.isdir(top_builddir):      sys.exit("didn't find top_builddir at {}".format(top_builddir))
if not os.path.isdir(top_srcdir):        sys.exit("didn't find top_srcdir at {}".format(top_srcdir))
if not os.path.isfile(tblfile):          sys.exit('pmark tbl file {} not found'.format(tblfile))
if not os.path.isfile(msafile):          sys.exit('pmark training MSA file {} not found'.format(msafile))
if not os.path.isfile(msafile + '.ssi'): sys.exit("msafile {} needs to have an SSI index.\nRun esl-afetch --index on it to create one.".format(msafile))
if not os.path.isfile(fafile):           sys.exit('pmark test sequence FASTA file {} not found'.format(fafile))
if not os.access(pmark_script, os.X_OK): sys.exit('driver script {} not found or not executable'.format(pmark_script))

os.mkdir(resultdir)
ncpu = int(ncpu)

# Read the master table, for MSAs with successful splits
#
msaname = []
alen    = {}
n       = 0
with open(tblfile) as tblfp:
    for line in tblfp:
        if line[0] == '#': continue   
        fields = line.split()

        if fields[7] == 'ok':
            msaname.append(fields[0])
            alen[fields[0]] = int(fields[2])
            n += 1

# Sort the list of MSA names by length of alignment (in columns),
# to help with load balancing; spread largest MSAs across the subtbls
#
msaname.sort(key=lambda s:alen[s], reverse=True)   


# Split the list into <ncpu> subtables
#
subtbl =[ [] for c in range(ncpu) ]
for i in range(n):
    subtbl[i % ncpu].append(msaname[i])

# Write the <ncpu> subtables
#
for c in range(ncpu):
    with open('{0}/tbl.{1}'.format(resultdir, c), 'w') as f:
        for s in subtbl[c]:
            f.write(s + '\n')
            
# Write a SLURM job array script
#
with open('{0}/{0}.sh'.format(resultdir), 'w') as f:
    cmd = '{0} {1} {2} {3} {3}/tbl.${{SLURM_ARRAY_TASK_ID}} {4} {5} {3}/tbl.${{SLURM_ARRAY_TASK_ID}}.out'.format(pmark_script, top_builddir, top_srcdir, resultdir, msafile, fafile)

    f.write('#!/bin/bash\n')
    f.write('#SBATCH -t 6-00:00\n')   # 6 days
    f.write('#SBATCH --mem 4000\n')
    f.write('#SBATCH -p eddy\n')
    f.write('#SBATCH -c 1\n')         # 1 core
    f.write('#SBATCH -N 1\n')         # 1 node
    f.write('#SBATCH -o {}/tbl.%a.slurm\n'.format(resultdir))
    f.write(cmd + '\n')

# Submit the job array
#
cmd = 'sbatch --array=0-{0} {1}/{1}.sh'.format(ncpu-1, resultdir)
subprocess.run(cmd.split())

    



# coding: utf-8
"""Test doctest contained tests in every file of the module.
"""

import configparser
import doctest
import importlib
import os
import pkgutil
import re
import sys
import shutil
import types
import warnings
from unittest import mock

try:
    import numpy
except ImportError:
    numpy = None

import pyhmmer.easel
import pyhmmer.plan7


def _load_tests_from_module(tests, module, globs, setUp=None, tearDown=None):
    """Load tests from module, iterating through submodules.
    """
    for attr in (getattr(module, x) for x in dir(module) if not x.startswith("_")):
        if isinstance(attr, types.ModuleType):
            suite = doctest.DocTestSuite(
                attr,
                globs,
                setUp=setUp,
                tearDown=tearDown,
                optionflags=+doctest.ELLIPSIS,
            )
            tests.addTests(suite)
    return tests


def load_tests(loader, tests, ignore):
    """`load_test` function used by unittest to find the doctests.
    """
    _current_cwd = os.getcwd()
    _daemon_client = mock.patch("pyhmmer.daemon.Client")

    # doctests are not compatible with `green`, so we may want to bail out
    # early if `green` is running the tests
    if sys.argv[0].endswith("green"):
        return tests

    # doctests require `numpy` to run, which may not be available because
    # it is a pain to get to work out-of-the-box on OSX inside CI
    if numpy is None:
        return tests

    # add a sample HMM and some sequences to use with the globals
    data = os.path.realpath(os.path.join(__file__, os.pardir, "data"))
    hmm_path = os.path.join(data, "hmms", "txt", "Thioesterase.hmm")
    with pyhmmer.plan7.HMMFile(hmm_path) as hmm_file:
        thioesterase = hmm_file.read()
    seq_path = os.path.join(data, "seqs", "938293.PRJEB85.HG003687.faa")
    with pyhmmer.easel.SequenceFile(seq_path, digital=True, alphabet=thioesterase.alphabet) as seq_file:
        proteins = seq_file.read_block()
    msa_path = os.path.join(data, "msa", "LuxC.faa")
    with pyhmmer.easel.MSAFile(msa_path, "afa") as msa_file:
        luxc = msa_file.read()
    seq_path = os.path.join(data, "seqs", "LuxC.faa")
    with pyhmmer.easel.SequenceFile(seq_path, digital=True) as seq_file:
        reductase = next(seq for seq in seq_file if b"P12748" in seq.name)

    def setUp(self):
        warnings.simplefilter("ignore")
        os.chdir(os.path.realpath(os.path.join(__file__, "..", "..")))
        # mock the HMMPGMD client to show usage examples without having
        # to actually spawn an HMMPGMD server in the background
        _client = _daemon_client.__enter__()
        _client.return_value = _client
        _client.__enter__.return_value = _client
        _client.connect.return_value = None
        _client.search_hmm.return_value = pyhmmer.plan7.TopHits(thioesterase)

    def tearDown(self):
        os.chdir(_current_cwd)
        warnings.simplefilter(warnings.defaultaction)
        _daemon_client.__exit__(None, None, None)

    # recursively traverse all library submodules and load tests from them
    packages = [None, pyhmmer]

    for pkg in iter(packages.pop, None):
        for (_, subpkgname, subispkg) in pkgutil.walk_packages(pkg.__path__):
            # do not import __main__ module to avoid side effects!
            if subpkgname == "__main__" or subpkgname.startswith("tests"):
                continue
            # import the submodule and add it to the tests
            module = importlib.import_module(".".join([pkg.__name__, subpkgname]))
            globs = dict(
                numpy=numpy,
                easel=pyhmmer.easel,
                plan7=pyhmmer.plan7,
                daemon=pyhmmer.daemon,
                thioesterase=thioesterase,
                proteins=proteins,
                luxc=luxc,
                reductase=reductase,
                pyhmmer=pyhmmer,
                **module.__dict__
            )
            tests.addTests(
                doctest.DocTestSuite(
                    module,
                    globs=globs,
                    setUp=setUp,
                    tearDown=tearDown,
                    optionflags=+doctest.ELLIPSIS,
                )
            )
            # if the submodule is a package, we need to process its submodules
            # as well, so we add it to the package queue
            if subispkg and subpkgname != "tests":
                packages.append(module)

    return tests



import typing
import functools

_T = typing.TypeVar("_T")
_S = typing.TypeVar("_S")


class peekable(typing.Iterator[_T], typing.Generic[_T]):
    """Turn an iterable into a peekable iterable.

    Example:
        >>> it = peekable(range(3))
        >>> it.peek()
        0
        >>> next(it)
        0
        >>> next(it)
        1
        >>> it.peek()
        2

    """

    _sentinel = object()

    def __init__(self, iterable: typing.Iterable[_T]):
        self.it = iter(iterable)
        self.peeked = self._sentinel

    def __iter__(self) -> "peekable[_T]":
        return self

    def __next__(self) -> _T:
        if self.peeked is not self._sentinel:
            self.peeked, item = self._sentinel, self.peeked
        else:
            item = next(self.it)
        return item  # type: ignore

    def peek(self) -> _T:
        if self.peeked is self._sentinel:
            self.peeked = next(self.it)
        return self.peeked  # type: ignore


class singledispatchmethod(typing.Generic[_T]):
    """Single-dispatch generic method descriptor.

    Backported from Python 3.8 to support older Python versions. Additional
    type annotations were obtained from the `functools` stubs in the Python
    `typeshed <https://github.com/python/typeshed>`_ repository.

    """

    def __init__(self, func: typing.Callable[..., _T]) -> None:
        if not callable(func) and not hasattr(func, "__get__"):
            raise TypeError("{!r} is not callable or a descriptor".format(func))
        self.dispatcher = functools.singledispatch(func)
        self.func = func

    @property
    def __isabstractmethod__(self) -> bool:
        return getattr(self.func, "__isabstractmethod__", False)

    def __get__(
        self, obj: _S, cls: typing.Optional[typing.Type[_S]]
    ) -> typing.Callable[..., _T]:
        def _method(*args, **kwargs):  # type: ignore
            method = self.dispatcher.dispatch(args[0].__class__)
            return method.__get__(obj, cls)(*args, **kwargs)

        _method.__isabstractmethod__ = self.__isabstractmethod__  # type: ignore
        _method.register = self.register  # type: ignore
        functools.update_wrapper(_method, self.func)
        return _method

    @typing.overload
    def register(  # type: ignore
        self, 
        cls: typing.Type[typing.Any], 
        method: None = ...
    ) -> typing.Callable[[typing.Callable[..., _T]], typing.Callable[..., _T]]:
        ...

    @typing.overload
    def register(  # type: ignore
        self, 
        cls: typing.Callable[..., _T], 
        method: None = ...
    ) -> typing.Callable[..., _T]:
        ...

    @typing.overload
    def register(  # type: ignore
        self, 
        cls: typing.Type[typing.Any],
        method: typing.Callable[..., _T]
    ) -> typing.Callable[..., _T]:
        ...

    def register(
        self, 
        cls: typing.Any, 
        method: typing.Optional[typing.Callable[..., _T]] = None
    ) -> typing.Any:
        """Registers a new implementation for the given class."""
        return self.dispatcher.register(cls, func=method)


class SizedIterator(typing.Generic[_T], typing.Iterator[_T], typing.Sized):
    """An iterator with a known number of items.
    """

    def __init__(self, n: int, iterable: typing.Iterable[_T]) -> None:
        self._it = iter(iterable)
        self._n = n

    def __len__(self) -> int:
        return self._n

    def __iter__(self: _S) -> _S:
        return self

    def __next__(self) -> _T:
        if self._n > 0:
            self._n -= 1
        return next(self._it)


from libc.stdio cimport FILE

from libeasel.msa cimport ESL_MSA
from libeasel.msafile cimport ESL_MSAFILE


cdef extern from "esl_msafile_a2m.h" nogil:
    int esl_msafile_a2m_SetInmap     (ESL_MSAFILE *afp)
    int esl_msafile_a2m_GuessAlphabet(ESL_MSAFILE *afp, int *ret_type)
    int esl_msafile_a2m_Read         (ESL_MSAFILE *afp, ESL_MSA **ret_msa)
    int esl_msafile_a2m_Write        (FILE *fp,    const ESL_MSA *msa)



from libc.stdint cimport uint64_t


cdef extern from "esl_bitfield.h" nogil:

    ctypedef struct ESL_BITFIELD:
        uint64_t* b
        int nb

    void esl_bitfield_Set(ESL_BITFIELD *b, int i)
    void esl_bitfield_Clear(ESL_BITFIELD *b, int i)
    void esl_bitfield_Toggle(ESL_BITFIELD *b, int i)
    bint esl_bitfield_IsSet(const ESL_BITFIELD *b, int i)

    ESL_BITFIELD* esl_bitfield_Create(int nb)
    int esl_bitfield_Count(const ESL_BITFIELD *b)
    void esl_bitfield_Destroy(ESL_BITFIELD *b)



from libc.stdint cimport uint32_t, int32_t, int64_t

cdef extern from "esl_dsqdata.h" nogil:

    const size_t eslDSQDATA_CHUNK_MAXSEQ
    const size_t eslDSQDATA_CHUNK_MAXPACKET
    const size_t eslDSQDATA_UNPACKERS
    const size_t eslDSQDATA_UMAX

    ctypedef esl_dsqdata_chunk_s ESL_DSQDATA_CHUNK
    cdef struct esl_dsqdata_chunk_s:
        int64_t i0
        int N

        ESL_DSQ** dsq
        char** name
        char** acc
        char** desc
        int32_t* taxid
        int64_t* L

        unsigned char* smem
        uint32_t *psq
        int pn
        char* metadata
        int mdalloc
        esl_dsqdata_chunk_s* nxt


    ctypedef esl_dsqdata_record_s ESL_DSQDATA_RECORD
    cdef struct esl_dsqdata_record_s:
        int64_t metadata_end
        int64_t psq_end


    ctypedef esl_dsqdata_s ESL_DSQDATA
    cdef esl_dsqdata_s:
        char* basename
        FILE* stubfp
        FILE* ifp
        FILE* sfp
        FILE* mfp
        ESL_ALPHABET* abc_r

        uint32_t magic
        uint32_t uniquetag
        uint32_t flags
        uint32_t max_namelen
        uint32_t max_acclen
        uint32_t max_desclen
        uint32_t max_seqlen
        uint32_t nseq
        uint32_t nres

        int chunk_maxseq
        int chunk_maxpacket
        int do_byteswap
        int pack5

        int nconsumers
        int n_unpackers

        ESL_DSQDATA_CHUNK*[eslDSQDATA_UMAX] inbox
        pthread_mutex_t[eslDSQDATA_UMAX] outbox_mutex
        pthread_cond_t[eslDSQDATA_UMAX] outbox_cv
        int[eslDSQDATA_UMAX] outbox_eod

        int64_t nchunk
        pthread_mutex_t nchunk_mutex

        ESL_DSQDATA_CHUNK* recycling
        pthread_mutex_t recycling_mutex
        pthread_cond_t recycling_cv

        int go
        pthread_mutex_t go_mutex
        pthread_cond_t go_cv

        pthread_t loader_t
        pthread_t unpacker_t[eslDSQDATA_UMAX]

        char[eslERRBUFSIZE] errbuf


    int  esl_dsqdata_Open   (ESL_ALPHABET **byp_abc, char *basename, int nconsumers, ESL_DSQDATA **ret_dd)
    int  esl_dsqdata_Read   (ESL_DSQDATA *dd, ESL_DSQDATA_CHUNK **ret_chu)
    int  esl_dsqdata_Recycle(ESL_DSQDATA *dd, ESL_DSQDATA_CHUNK *chu)
    int  esl_dsqdata_Close  (ESL_DSQDATA *dd)
    int  esl_dsqdata_Write  (ESL_SQFILE *sqfp, char *basename, char *errbuf)



#! /usr/bin/env python3

# autodoc.py
#
# Extract function documentation from a .c file and output it in a
# Markdown file. Assumes that the .c file uses function headers that
# conform to Easel code style. 
# 
# Usage:    ./autodoc.py <.c file>
# Example:  ./autodoc.py esl_module.c > esl_module_functions.md
#
#
# Replaces old Perl autodoc, which converted to LaTeX. I'm moving
# toward online developer documentation, and Markdown. This simplifies
# the conversion; we don't have to protect special LaTeX characters
# (#_$) for example.
#
# SRE, Sun 27 Jan 2019

import os
import sys
import re
import getopt

def process(text):
    """
    Remove the leading " *     " prefixes from a Purpose, Returns, or Throws
    multiiline block of text that we've just pulled out of the function header.
    
    Convert <text> to `text` (Markdown code).
    """
    text = re.sub(r'^ \*[ \t]+(\S)', r'\1',   text, flags=re.MULTILINE)               # remove leading " *   "
    text = re.sub(r'^ \*',           r'',     text, flags=re.MULTILINE)               # remove " *" alone
    text = re.sub(r'<(\S|\S.*?\S)>', r'`\1`', text, flags=re.MULTILINE | re.DOTALL)   # convert <text> to `text`
    return text

def output_argtable(argtext):
    print("|  arg | description |")
    print("|------|-------------|")
    for m in re.finditer(r'^(\S+)\s*[:-]\s*(.+?)\s*$', argtext, flags=re.MULTILINE):
        print("| `{0}` | {1} |".format(m.group(1), m.group(2)))
    print("\n")

def main():
    try:    (opts, args) = getopt.getopt(sys.argv[1:], "t")
    except:              sys.exit("Usage: autodoc.py [-t] <.c file>")
    if (len(args) != 1): sys.exit("Usage: autodoc.py [-t] <.c file>")        

    cfile    = args[0]
    do_table = False
    for opt, arg in opts:
        if opt == '-t': do_table = True
    
    if not os.path.isfile(cfile): exit(".c file {0} not found".format(cfile))
    fp   = open(cfile)
    text = fp.read()

    if do_table:
        print('| {0:30s} | {1:60s} |'.format('Function', 'Synopsis'))
        print('|{0:-^32s}|{1:-^62s}|'.format('', ''))
              

    #                        /* Function:  ...    */   ... } to a blank line.   Grabs header + implementation(s)...
    #                        vv    vvvvvvvv       vv   vvv   vvv   v...      or, grabs subheading line in a comment
    for m in re.finditer(r'^(/\*\s+Function:.+?^ \*/)(.+?^\})\s*$^\s*$|^\s*\*#\s*\d+\..+?$\s*', text, flags=re.MULTILINE | re.DOTALL):
        if m.group(0).startswith('/*'):
            header = m.group(1)    # comment header "/* Function: ... */"
            impl   = m.group(2)    # implementation(s) "int myfunc(args){  }\nint func2(args){ }"

            m = re.match(r'/\*\s+Function:\s*(.+)$', header, flags=re.MULTILINE)    # Usually one function name, but could also be comma-delimited list    
            funcnames = [ a.lstrip().rstrip() for a in m.group(1).split(',') ]    

            m = re.search(r'^\s+\*\s+Synopsis:\s+(.+)$', header, flags=re.MULTILINE)
            synopsis = process(m.group(1)) if m else None

            m = re.search(r'^\s+\*\s+Args:\s+(.+?)(?:^ \*/|^ \* \S)', header, flags=re.MULTILINE | re.DOTALL)
            argtext = process(m.group(1)) if m else None

            m = re.search(r'^\s+\*\s+Purpose:\s+(.+?)(?:^ \*/|^ \* \S)', header, flags=re.MULTILINE | re.DOTALL)
            purpose = process(m.group(1)) if m else None

            m = re.search(r'^\s+\*\s+Returns:\s+(.+?)(?:^ \*/|^ \* \S)', header, flags=re.MULTILINE | re.DOTALL)
            returns = process(m.group(1)) if m else None

            m = re.search(r'^\s+\*\s+Throws:\s+(.+?)(?:^ \*/|^ \* \S)', header, flags=re.MULTILINE | re.DOTALL)
            throws = process(m.group(1)) if m else None

            # pull the call syntax (function name, arguments) out of the C implementation;
            # <syntax> is a list of each documented function and its call syntax, "int foo(double bar)".
            # nontrivial to do well with just regexps, without a real grammar parser,
            # because we're covering the less common case where there's >1 function
            # documented by a single header.
            syntax = []
            for fname in funcnames:                                 # list of names like "esl_foo_Function()", with the (). Can also be esl_foo_{DFI}Function()", which needs expansion.
                fname = fname.rstrip('()')                          # now just "esl_foo_Function" or "esl_foo_{DFI}Function()"
                m     = re.search(r'\{([DFILCWB]+)\}', fname)       # "esl_foo_{DFI}Function()" case?
                if m:                                               #   then expand it, one function name at a time
                    typelist = m.group(1)                           # "DFI" for example
                    for c in typelist:
                        expanded_fname = re.sub(r'\{([DFILCWB]+)\}', c, fname)
                        impl_pattern   = r'^(\S+.+?)\s+' + expanded_fname + '\s*(\((?s:.+?)\))\s*\{'
                        m = re.search(impl_pattern, impl, flags=re.MULTILINE)
                        if m: syntax.append(m.group(1) + ' ' + expanded_fname + m.group(2))
                        else: exit("failed to parse out the syntax for {0}".format(expanded_fname))
                else:
                    pattern1 = r'^(\S+.+?)\s+' + fname + '\s*(\((?s:.+?)\))\s*\{'
                    m = re.search(pattern1, impl, flags=re.MULTILINE)
                    if m: syntax.append(m.group(1) + ' ' + fname + m.group(2))
                    else: exit("failed to parse out the syntax for {0}".format(fname))

            # Now we're done parsing one (or more) documented functions,
            # and it's time to print whatever we're going to print.
            #
            #   funcnames : list of one (or more) function names sharing the documentation
            #   synopsis  : optional one-line description
            #   syntax    : list of one (or more) "funcname(arg, arg)" call syntax
            #   purpose   : optional documentation (Markdown format)
            #   argtext   : optional argument table text (needs further processing)
            #   returns:  : optional text about return status
            #   throws:   : optional text about exceptions
            #
            if do_table:
                if synopsis: print('| {0:30s} | {1:60s} |'.format('`{}`'.format(funcnames[0]), synopsis))
                else:        print('| {0:30s} | {1:60s} |'.format('`{}`'.format(funcnames[0]), ''))
            else:
                for a in funcnames: print("### `{0}`\n".format(a))
                if synopsis: print("**{0}**\n".format(synopsis.rstrip())) 
                for s in syntax:    print("`{0}`\n".format(s))
                if argtext: output_argtable(argtext)
                if purpose: print(purpose);
                if returns: print("Returns: {0}".format(returns))
                if throws:  print("Throws: {0}".format(throws))
                print("------")

        else:  # or, we're a section heading.
            m = re.match('^\s*\*#\s*(\d+\..+)', m.group(0))
            secheading = m.group(1)

            if not do_table:
                print("## {0}\n".format(secheading))


if __name__ == "__main__":
    main()







    




